import numpy as np
import uproot
import re
import pandas as pd

import lhcb_rex.processing.transformers as tfs
import lhcb_rex.tools.variables_tools as pts
from lhcb_rex.rapidsim.rapidsim_tools import compute_combined_param
import lhcb_rex.settings.globals as myGlobals
import lhcb_rex.tools.display as display

masses = {}
masses[321] = 493.677
masses[211] = 139.57039
masses[13] = 105.66
masses[11] = 0.51099895000  # * 1e-3
pid_list = [11, 13, 211, 321]


rapidsim_conventions = {}

rapidsim_conventions["momenta_component"] = "{particle}_P"
rapidsim_conventions["true_momenta_component"] = "{particle}_P_TRUE"

rapidsim_conventions["momenta_component"] = "{particle}_P{dim}"
rapidsim_conventions["true_momenta_component"] = "{particle}_P{dim}_TRUE"

rapidsim_conventions["pid"] = "{particle}_ID"
rapidsim_conventions["true_pid"] = "{particle}_ID_TRUE"

rapidsim_conventions["mass"] = "{particle}_M"
rapidsim_conventions["true_mass"] = "{particle}_M_TRUE"

rapidsim_conventions["origin"] = "{particle}_orig{dim}"
rapidsim_conventions["true_origin"] = "{particle}_orig{dim}_TRUE"

rapidsim_conventions["vertex"] = "{particle}_vtx{dim}"
rapidsim_conventions["true_vertex"] = "{particle}_vtx{dim}_TRUE"


class tuple_manager:
    def map_branch_names(self):
        branch_names = list(self.tuple.keys())
        branch_names = [
            branch.replace(self.mother_particle_name, self.mother)
            for branch in branch_names
        ]
        # if self.intermediate_particle_name:
        #     branch_names = [
        #         branch.replace(self.intermediate_particle_name, self.intermediate)
        #         for branch in branch_names
        #     ]
        temps = []
        for i in range(self.Nparticles):  # Nparticles
            if (
                self.particles[i] in self.daughter_particle_names
                and self.particles[i] == self.daughter_particle_names[i]
            ):
                # all good - no need really even to replace
                replace_string = self.particles[i]
            else:
                replace_string = f"{self.particles[i][0]}_temp_{self.particles[i][1:]}"
                temps.append(i)
            branch_names = [
                branch.replace(self.daughter_particle_names[i], replace_string)
                for branch in branch_names
            ]
        for i in temps:
            replace_string = f"{self.particles[i][0]}_temp_{self.particles[i][1:]}"
            branch_names = [
                branch.replace(replace_string, self.particles[i])
                for branch in branch_names
            ]
        self.tuple.columns = branch_names

    def map_branch_names_list(self, branch_names):
        branch_names = [
            branch.replace(self.mother_particle_name, self.mother)
            for branch in branch_names
        ]
        # if self.intermediate_particle_name:
        #     branch_names = [
        #         branch.replace(self.intermediate_particle_name, self.intermediate)
        #         for branch in branch_names
        #     ]

        print(myGlobals.particle_map["toDAUGHTERS"].keys())
        print(myGlobals.particle_map["fromDAUGHTERS"].keys())
        #
        quit()

        temps = []
        for i in range(self.Nparticles):  # Nparticles
            if (
                self.particles[i] in self.daughter_particle_names
                and self.particles[i] == self.daughter_particle_names[i]
            ):
                # all good - no need really even to replace
                replace_string = self.particles[i]
            else:
                replace_string = f"{self.particles[i][0]}_temp_{self.particles[i][1:]}"
                temps.append(i)
            branch_names = [
                branch.replace(self.daughter_particle_names[i], replace_string)
                for branch in branch_names
            ]
        for i in temps:
            replace_string = f"{self.particles[i][0]}_temp_{self.particles[i][1:]}"
            branch_names = [
                branch.replace(replace_string, self.particles[i])
                for branch in branch_names
            ]
        return branch_names

    # def smearelectronE(self, E_smearing_output, particle):
    #     for column in E_smearing_output.columns:
    #         dim = column[-1]
    #         self.tuple[f"{particle}_P{dim}"] = (
    #             E_smearing_output[column]
    #             * (self.tuple[f"{particle}_P{dim}_TRUE"] + 1e-4)
    #         ) + self.tuple[f"{particle}_P{dim}_TRUE"]

    # def propagate_smearing(self, smearing_output, Nparticles=3):

    #     for idx in range(Nparticles):
    #         for dim in ["X", "Y"]:
    #             self.tuple[f"DAUGHTER{idx + 1}_P{dim}"] = (
    #                 smearing_output[f"DAUGHTER{idx + 1}_delta_P{dim}"]
    #                 + self.tuple[f"DAUGHTER{idx + 1}_P{dim}_TRUE"]
    #             )

    #     for idx in range(Nparticles):
    #         true_P = np.sqrt(
    #             self.tuple[f"DAUGHTER{idx + 1}_PX_TRUE"] ** 2
    #             + self.tuple[f"DAUGHTER{idx + 1}_PY_TRUE"] ** 2
    #             + self.tuple[f"DAUGHTER{idx + 1}_PZ_TRUE"] ** 2
    #         )
    #         A_sq = smearing_output[f"DAUGHTER{idx + 1}_Preco_overP"] ** 2
    #         self.tuple[f"DAUGHTER{idx + 1}_PZ"] = np.sqrt(
    #             A_sq * true_P**2
    #             - self.tuple[f"DAUGHTER{idx + 1}_PX"] ** 2
    #             - self.tuple[f"DAUGHTER{idx + 1}_PY"] ** 2
    #         )

    #     for idx in range(Nparticles):
    #         self.tuple[f"DAUGHTER{idx + 1}_TRACK_CHI2NDOF"] = smearing_output[f"DAUGHTER{idx + 1}_TRACK_CHI2NDOF"]
    #         self.tuple[f"DAUGHTER{idx + 1}_TRACK_GhostProb"] = smearing_output[f"DAUGHTER{idx + 1}_TRACK_GhostProb"]

    def propagate_smearing(self, smearing_output, Nparticles=3):
        for idx in range(Nparticles):
            for dim in ["X", "Y", "Z"]:
                self.tuple[f"DAUGHTER{idx + 1}_P{dim}"] = (
                    smearing_output[f"DAUGHTER{idx + 1}_delta_P{dim}"]
                    + self.tuple[f"DAUGHTER{idx + 1}_P{dim}_TRUE"]
                )

        for idx in range(Nparticles):
            self.tuple[f"DAUGHTER{idx + 1}_TRACK_CHI2NDOF"] = smearing_output[
                f"DAUGHTER{idx + 1}_TRACK_CHI2NDOF"
            ]
            self.tuple[f"DAUGHTER{idx + 1}_TRACK_GhostProb"] = smearing_output[
                f"DAUGHTER{idx + 1}_TRACK_GhostProb"
            ]

        # smearing_output[f"DAUGHTER{idx + 1}_Preco_overP"]

        for idx in range(Nparticles):
            self.tuple[f"DAUGHTER{idx + 1}_P"] = np.sqrt(
                (
                    self.tuple[f"DAUGHTER{idx + 1}_PX"] ** 2
                    + self.tuple[f"DAUGHTER{idx + 1}_PY"] ** 2
                    + self.tuple[f"DAUGHTER{idx + 1}_PZ"] ** 2
                )
            )
            self.tuple[f"DAUGHTER{idx + 1}_TRUEP"] = np.sqrt(
                (
                    self.tuple[f"DAUGHTER{idx + 1}_PX_TRUE"] ** 2
                    + self.tuple[f"DAUGHTER{idx + 1}_PY_TRUE"] ** 2
                    + self.tuple[f"DAUGHTER{idx + 1}_PZ_TRUE"] ** 2
                )
            )
            self.tuple[f"DAUGHTER{idx + 1}_Preco_overP"] = (
                self.tuple[f"DAUGHTER{idx + 1}_P"]
                / self.tuple[f"DAUGHTER{idx + 1}_TRUEP"]
            )

    def recompute_combined_particles(self, combined_particles, daughter_particle_names):
        # Recompute variables of reconstructed particles

        for combined_particle in combined_particles:
            daughter_info = {}
            for daughter in combined_particles[combined_particle]:
                daughter_info[daughter] = {}
                for param in ["M", "PX", "PY", "PZ"]:
                    daughter_info[daughter][param] = self.tuple[f"{daughter}_{param}"]
            for param in ["M", "PX", "PY", "PZ"]:
                self.tuple[f"{combined_particle}_{param}"] = compute_combined_param(
                    daughter_info, param
                )
            self.tuple[f"{combined_particle}_P"] = np.sqrt(
                self.tuple[f"{combined_particle}_PX"] ** 2
                + self.tuple[f"{combined_particle}_PY"] ** 2
                + self.tuple[f"{combined_particle}_PZ"] ** 2
            )
            self.tuple[f"{combined_particle}_PT"] = np.sqrt(
                self.tuple[f"{combined_particle}_PX"] ** 2
                + self.tuple[f"{combined_particle}_PY"] ** 2
            )

    def recompute_reconstructed_mass(self):
        df = self.tuple.copy()

        for _idx, particle in enumerate(self.particles):
            if _idx == 0:
                PE = np.sqrt(
                    df[f"{particle}_M"] ** 2
                    + df[f"{particle}_PX"] ** 2
                    + df[f"{particle}_PY"] ** 2
                    + df[f"{particle}_PZ"] ** 2
                )
                PX = df[f"{particle}_PX"]
                PY = df[f"{particle}_PY"]
                PZ = df[f"{particle}_PZ"]
            else:
                PE += np.sqrt(
                    df[f"{particle}_M"] ** 2
                    + df[f"{particle}_PX"] ** 2
                    + df[f"{particle}_PY"] ** 2
                    + df[f"{particle}_PZ"] ** 2
                )
                PX += df[f"{particle}_PX"]
                PY += df[f"{particle}_PY"]
                PZ += df[f"{particle}_PZ"]

        mass = np.sqrt((PE**2 - PX**2 - PY**2 - PZ**2))

        return mass

    def ensure_consistent_final_state(self, tag=""):
        indices_to_drop = set()  # Store indices to drop later

        true_PID_scheme = {}
        for idx, particle in enumerate(self.particles):
            masses = np.ones_like(self.tuple[f"{tag}{particle}_ID_TRUE"])
            PIDs_present = []

            for PID in myGlobals.particle_universe:
                where = np.where(
                    np.abs(self.tuple[f"{tag}{particle}_ID_TRUE"].astype(int)) == PID
                )
                masses[where] = myGlobals.masses[PID]
                if np.shape(where)[1] > 0:
                    PIDs_present.append(PID)

            if len(PIDs_present) != 1:
                display.warning_print(PIDs_present)
                PIDs_present, counts = np.unique(
                    np.abs(self.tuple[f"{tag}{particle}_ID_TRUE"].astype(int)),
                    return_counts=True,
                )
                where = np.where(counts == np.amax(counts))
                select = PIDs_present[where][0]

                display.warning_print(
                    "WARNING: particle type cannot change, TRUEID for each final state particle must be the same throughout the tuple."
                )
                display.warning_print(
                    f"present in {particle} {PIDs_present} (counts: {counts})"
                )
                display.warning_print(
                    f"Marking rows for removal where {tag}{particle}_ID_TRUE != {select}"
                )

                # Identify indices to drop
                drop_indices = self.tuple.index[
                    np.abs(self.tuple[f"{tag}{particle}_ID_TRUE"]) != select
                ].tolist()
                indices_to_drop.update(drop_indices)

            true_PID_scheme[particle] = np.abs(
                self.tuple[f"{tag}{particle}_ID_TRUE"].astype(int)
            )[0]

        # Drop rows at the end
        if indices_to_drop:
            display.warning_print(
                f"Dropping {len(indices_to_drop)} rows at the end of processing."
            )
            self.tuple = self.tuple.drop(index=indices_to_drop).reset_index(drop=True)

        return true_PID_scheme

    def __init__(
        self,
        tuple_location,
        mother_particle_name,
        daughter_particle_names,
        combined_particles,
        tree="DecayTree",
        branch_naming_structure=None,
        mass_hypotheses=None,
    ):
        # print("Class `tuple_manager` initialised with the following inputs:")
        # for arg_name, arg_value in locals().items():
        #     print(f"  {arg_name} = {arg_value}")

        self.mother = "MOTHER"

        raw_tuple = uproot.open(tuple_location)[tree]
        self.tuple_location = tuple_location
        list_of_branches = list(raw_tuple.keys())
        list_of_branches = [
            branch
            for branch in list_of_branches
            if "COV" not in branch and branch != "index"
        ]
        arrays = raw_tuple.arrays(list_of_branches, library="np")
        self.tuple = pd.DataFrame(arrays)
        if branch_naming_structure:
            self.convert_branch_aliases(branch_naming_structure, to_rapidsim=True)

        self.original_branches = list(self.tuple.keys())

        # # code to replace names
        # new_branches = []
        # for branch in self.tuple:
        #     branch.replace(mother_particle_name,'MOTHER')
        #     for combined_particle in combined_particles:
        #         if combined_particle == "MOTHER":
        #             branch = branch.replace(mother_particle_name,combined_particle)
        #         else:
        #             branch = branch.replace(myGlobals.particle_map['fromINTERMEDIATES'][combined_particle],combined_particle)
        #     for particle in list(myGlobals.particle_map['toDAUGHTERS'].keys()):
        #         branch = branch.replace(particle, myGlobals.particle_map['toDAUGHTERS'][particle])
        #     new_branches.append(branch)
        # self.tuple.columns = new_branches

        # code to replace names
        new_branches = []
        temp_particles = {}
        for idx, particle in enumerate(
            list(myGlobals.particle_map["toDAUGHTERS"].keys())
        ):
            temp_particles[particle] = str(np.random.randint(0, 99999999))

        for branch in self.tuple:
            branch.replace(mother_particle_name, "MOTHER")
            for combined_particle in combined_particles:
                if combined_particle == "MOTHER":
                    branch = branch.replace(mother_particle_name, combined_particle)
                else:
                    branch = branch.replace(
                        myGlobals.particle_map["fromINTERMEDIATES"][combined_particle],
                        combined_particle,
                    )
            for particle in list(myGlobals.particle_map["toDAUGHTERS"].keys()):
                branch = branch.replace(particle, temp_particles[particle])
            for particle in list(myGlobals.particle_map["toDAUGHTERS"].keys()):
                branch = branch.replace(
                    temp_particles[particle],
                    myGlobals.particle_map["toDAUGHTERS"][particle],
                )
            new_branches.append(branch)
        # for new_branche in new_branches:
        #     print(new_branche)
        # quit()
        self.tuple.columns = new_branches

        self.particles = [
            myGlobals.particle_map["toDAUGHTERS"][daughter]
            for daughter in daughter_particle_names
        ]

        for _idx, particle in enumerate(self.particles):
            if f"{particle}_M" not in self.tuple:
                if mass_hypotheses is not None:
                    if particle in mass_hypotheses:
                        if (
                            myGlobals.rapidsim_settings["RAPIDSIM_particles_df"][
                                (
                                    myGlobals.rapidsim_settings[
                                        "RAPIDSIM_particles_df"
                                    ]["part"]
                                    == mass_hypotheses[particle]
                                )
                            ].shape[0]
                            == 1
                        ):
                            pdg = int(
                                myGlobals.rapidsim_settings["RAPIDSIM_particles_df"][
                                    (
                                        myGlobals.rapidsim_settings[
                                            "RAPIDSIM_particles_df"
                                        ]["part"]
                                        == mass_hypotheses[particle]
                                    )
                                ].ID
                            )
                        elif (
                            myGlobals.rapidsim_settings["RAPIDSIM_particles_df"][
                                (
                                    myGlobals.rapidsim_settings[
                                        "RAPIDSIM_particles_df"
                                    ]["anti"]
                                    == mass_hypotheses[particle]
                                )
                            ].shape[0]
                            == 1
                        ):
                            pdg = int(
                                myGlobals.rapidsim_settings["RAPIDSIM_particles_df"][
                                    (
                                        myGlobals.rapidsim_settings[
                                            "RAPIDSIM_particles_df"
                                        ]["anti"]
                                        == mass_hypotheses[particle]
                                    )
                                ].ID
                                * -1
                            )

                        pdg = int(pdg)
                    else:
                        pdg = int(self.tuple[f"{particle}_ID"][0])
                else:
                    pdg = int(self.tuple[f"{particle}_ID"][0])

                self.tuple[f"{particle}_M"] = masses[abs(pdg)]

            if f"{particle}_ID_TRUE" not in self.tuple:
                self.tuple[f"{particle}_ID_TRUE"] = self.tuple[f"{particle}_ID"]
                if mass_hypotheses is not None:
                    if particle in mass_hypotheses:
                        self.tuple[f"{particle}_ID"] = pdg

        self.true_PID_scheme = self.ensure_consistent_final_state()
        self.recompute_combined_particles(combined_particles, daughter_particle_names)

    def convert_physical_units(self, conversion, specific_tuple=None):
        if conversion not in ["from_MeV", "back_to_MeV"]:
            print("conversion not valid")
            raise

        if specific_tuple is not None:
            tuple_to_update = specific_tuple
        else:
            tuple_to_update = self.tuple

        # do this before and after
        if conversion == "from_MeV":
            self.branches_for_conversion = []
            for branch in tuple_to_update:
                for (
                    branch_naming_structure,
                    pattern,
                ) in rapidsim_conventions.items():
                    if branch_naming_structure in [
                        "momenta_component",
                        "true_momenta_component",
                        "mass",
                        "true_mass",
                    ]:
                        regex_pattern = (
                            pattern.replace("{particle}", r"(?P<particle>\w+)").replace(
                                "{dim}", r"(?P<dim>\w)"
                            )
                            + r"$"
                        )  # Only a single character for {dim}
                        match = re.match(regex_pattern, branch)

                        if match:
                            self.branches_for_conversion.append(branch)

            for branch in self.branches_for_conversion:
                # tuple_to_update.loc[:, branch] *= 1e-3
                tuple_to_update.loc[:, branch] = tuple_to_update.loc[:, branch] * 1e-3
        else:
            for branch in self.branches_for_conversion:
                # tuple_to_update.loc[:, branch] *= 1e3
                tuple_to_update.loc[:, branch] = tuple_to_update.loc[:, branch] * 1e3

    def convert_branch_aliases(
        self, branch_naming_structures, to_rapidsim, specific_tuple=None
    ):
        if specific_tuple is not None:
            tuple_to_update = specific_tuple
        else:
            tuple_to_update = self.tuple

        if to_rapidsim:
            new_branches = []
            for branch in tuple_to_update:
                new_branch = branch

                for (
                    branch_naming_structure,
                    pattern,
                ) in branch_naming_structures.items():
                    regex_pattern = (
                        pattern.replace("{particle}", r"(?P<particle>\w+)").replace(
                            "{dim}", r"(?P<dim>\w)"
                        )
                        + r"$"
                    )  # Only a single character for {dim}
                    match = re.match(regex_pattern, branch)

                    if match:
                        particle = match.group("particle")
                        try:
                            dim = match.group("dim")
                        except Exception:
                            dim = ""  # no dim, for example mass branch
                        updated_branch = rapidsim_conventions[
                            branch_naming_structure
                        ].format(particle=particle, dim=dim)
                        new_branch = updated_branch
                        break
                new_branches.append(new_branch)

            tuple_to_update.columns = new_branches
        else:
            new_branches = []

            for branch in tuple_to_update:
                new_branch = branch

                for (
                    branch_naming_structure,
                    custom_pattern,
                ) in branch_naming_structures.items():
                    pattern = rapidsim_conventions[branch_naming_structure]

                    regex_pattern = (
                        pattern.replace("{particle}", r"(?P<particle>\w+)").replace(
                            "{dim}", r"(?P<dim>\w)"
                        )
                        + r"$"
                    )  # Only a single character for {dim}
                    match = re.match(regex_pattern, branch)

                    if match:
                        particle = match.group("particle")
                        try:
                            dim = match.group("dim")
                        except Exception:
                            dim = ""  # no dim, for example mass branch
                        updated_branch = custom_pattern.format(
                            particle=particle, dim=dim
                        )
                        new_branch = updated_branch
                        break

                new_branches.append(new_branch)

            tuple_to_update.columns = new_branches

    def write(
        self,
        new_branches_to_keep,
        output_location=None,
        keep_vertex_info=False,
        keep_tuple_structure=False,
        extra_branches=[],
    ):
        branches = self.original_branches + new_branches_to_keep

        if not keep_tuple_structure:
            # re-name columns
            branch_swaps = {}
            branch_swaps[self.mother] = self.mother_particle_name
            if self.intermediate_particle_name:
                if isinstance(self.intermediate_particle_name, list):
                    for inter in self.intermediate_particle_name:
                        branch_swaps[inter] = inter
                else:
                    branch_swaps[self.intermediate_particle_name] = (
                        self.intermediate_particle_name
                    )
            branch_swaps[self.particles[0]] = self.daughter_particle_names[0]
            branch_swaps[self.particles[1]] = self.daughter_particle_names[1]
            branch_swaps[self.particles[2]] = self.daughter_particle_names[2]
            # add rest to list - will only be others if dropMissing=False
            if isinstance(self.intermediate_particle_name, list):
                named_particles = (
                    [self.mother] + self.intermediate_particle_name + self.particles
                )
            else:
                named_particles = (
                    [self.mother] + [self.intermediate_particle_name] + self.particles
                )
            unnamed_particles = list(set(self.list_of_particles) - set(named_particles))
            for unnamed_particle in unnamed_particles:
                branch_swaps[unnamed_particle] = unnamed_particle

        tuple_to_write = self.tuple[branches]

        if not keep_tuple_structure:
            columns = list(tuple_to_write.columns)
            new_columns = []
            for column in columns:
                for to_swap in list(branch_swaps.keys()):
                    if column[: len(to_swap)] == to_swap:
                        new_columns.append(
                            column.replace(to_swap, branch_swaps[to_swap])
                        )
                        break

            # drop columns that might hang on but are not related to an individual particle, MCorr and nEvent for example
            drop_list = []
            for i in tuple_to_write.columns:
                # if i not in new_columns:
                if not any(s in i for s in list(branch_swaps.keys())):
                    drop_list.append(i)
            tuple_to_write = tuple_to_write.drop(drop_list, axis=1)

            tuple_to_write.columns = new_columns

            if not keep_vertex_info:
                for dim in ["X", "Y", "Z"]:
                    tuple_to_write = tuple_to_write.drop(
                        columns=[
                            col for col in tuple_to_write.columns if f"_vtx{dim}" in col
                        ]
                    )
                    tuple_to_write = tuple_to_write.drop(
                        columns=[
                            col
                            for col in tuple_to_write.columns
                            if f"_orig{dim}" in col
                        ]
                    )

            # re-order columns
            columns = tuple_to_write.columns
            if self.intermediate_particle_name:
                prefix_order = [self.mother_particle_name]
                if not isinstance(self.intermediate_particle_name, list):
                    prefix_order.append(self.intermediate_particle_name)
                else:
                    prefix_order.extend(self.intermediate_particle_name)
                prefix_order.extend(
                    [
                        self.daughter_particle_names[0],
                        self.daughter_particle_names[1],
                        self.daughter_particle_names[2],
                    ]
                )
            else:
                prefix_order = [
                    self.mother_particle_name,
                    self.daughter_particle_names[0],
                    self.daughter_particle_names[1],
                    self.daughter_particle_names[2],
                ]

            for unnamed_particle in unnamed_particles:
                prefix_order.append(unnamed_particle)

            ordered_columns = []
            for prefix in prefix_order:
                cols_with_prefix = [col for col in columns if col.startswith(prefix)]
                ordered_columns.extend(cols_with_prefix)
            tuple_to_write = tuple_to_write[ordered_columns]

        # if self.physical_units != "GeV":
        #     self.convert_physical_units(
        #         conversion="back_to_MeV", specific_tuple=tuple_to_write
        #     )

        if self.branch_naming_structure:
            self.convert_branch_aliases(
                self.branch_naming_structure,
                to_rapidsim=False,
                specific_tuple=tuple_to_write,
            )

        if len(extra_branches) > 0:
            tuple_to_write = tuple_to_write.drop(
                columns=extra_branches, errors="ignore"
            )
            tuple_to_write = pd.concat(
                [tuple_to_write, self.tuple[extra_branches]], axis=1
            )

        if not output_location:
            output_location = f"{self.tuple_location[:-5]}_reco.root"
        pts.write_df_to_root(tuple_to_write, output_location, self.tree)
        return output_location

    def add_branches(self, data_to_add, append_to_leaf_vector=False):
        if append_to_leaf_vector:
            for branch, data in data_to_add.items():
                # If the branch already exists, append_to_leaf_vector
                if branch in self.tuple.columns:
                    current_data = np.asarray(self.tuple[branch])

                    current_data = np.vstack(current_data)

                    if len(np.shape(current_data)) == 1:
                        current_data = np.expand_dims(current_data, 1)
                    if len(np.shape(data)) == 1:
                        data = np.expand_dims(data, 1)
                    new_branch_vector = np.concatenate((current_data, data), axis=1)

                    self.tuple[branch] = [list(row) for row in new_branch_vector]

                else:
                    # If the branch doesn't exist, add it as a new column
                    self.tuple[branch] = pd.DataFrame({branch: data})
        else:
            # Initialize a dictionary to collect the new columns
            columns_i = {}

            # Loop over the items in `data_to_add`
            for branch, data in data_to_add.items():
                if branch in self.tuple.columns:
                    # If the branch already exists, overwrite the column with new data
                    self.tuple[branch + "_OLD"] = self.tuple[branch].copy()
                    self.tuple[branch] = data
                else:
                    # If the branch doesn't exist, add it to the dictionary for future concatenation
                    columns_i[branch] = data

            # After the loop, concatenate the new columns in `columns_i` to the existing DataFrame
            if columns_i:
                new_columns_df = pd.DataFrame(columns_i)
                self.tuple = pd.concat([self.tuple, new_columns_df], axis=1)

    def get_branches(
        self,
        branches,
        transformers=None,
        numpy=False,
        scale_factor=1.0,
        transform_by_index=False,
        tag="",
        external_tuple=None,
        change_units={},
    ):
        if external_tuple is not None:
            working_tuple = external_tuple
        else:
            working_tuple = self.tuple

        try:
            data = working_tuple[branches] * scale_factor
        except Exception as e:
            for branch in branches:
                try:
                    working_tuple[branch]
                except Exception as e:
                    print(branch, f"not there {e}")
            print(f"quitting... {e}")
            raise Exception("branches not in working_tuple")

        for item in change_units:
            data[item] *= change_units[item]

        if transformers:
            data = tfs.transform_df(
                data, transformers, transform_by_index=transform_by_index, tag=tag
            )

        if numpy:
            data = np.asarray(data[branches])

        return data

    def get_condition_chunks(
        self, network, particles_involved, batch_size, name="MOTHER", mother=True
    ):
        plot_conditions = False
        if plot_conditions:
            save_cond_info = {}
            save_cond_info["processed"] = {}
            save_cond_info["physical"] = {}

        Nparticles = len(network.branch_options["particles_involved"][0])

        self.tuple["N_daughters"] = Nparticles

        conditions_graph = []
        personalised_transformers = {}
        for condition in network.branch_options["mother_conditions"]:
            conditions_graph.append(condition)
            if condition != "N_daughters":
                personalised_transformers[condition] = network.Transformers[condition]
        mother_conditions = self.get_branches(
            conditions_graph,
            personalised_transformers,
            numpy=True,
            # numpy=False,
        )

        conditions_graph = []
        personalised_transformers = {}
        for condition in network.branch_options["intermediate_conditions"]:
            conditions_graph.append(condition)
            if condition != "N_daughters":
                personalised_transformers[condition] = network.Transformers[condition]
        intermediate_conditions = self.get_branches(
            conditions_graph,
            personalised_transformers,
            numpy=True,
            # numpy=False,
        )

        conditions_graph = []
        personalised_transformers = {}
        for condition in network.branch_options["track_conditions"]:
            for i in range(Nparticles):
                conditions_graph.append(
                    condition.replace("DAUGHTERN", f"DAUGHTER{i + 1}")
                )
                if condition != "N_daughters":
                    personalised_transformers[
                        condition.replace("DAUGHTERN", f"DAUGHTER{i + 1}")
                    ] = network.Transformers[
                        condition.replace("DAUGHTERN", "DAUGHTER1")
                    ]
        track_conditions = self.get_branches(
            conditions_graph,
            personalised_transformers,
            numpy=True,
            # numpy=False,
        )
        track_conditions = np.reshape(
            track_conditions, (np.shape(track_conditions)[0], Nparticles, -1)
        )

        mother_chunks = [
            mother_conditions[i : i + batch_size]
            for i in range(0, len(mother_conditions), batch_size)
        ]
        intermediate_chunks = [
            intermediate_conditions[i : i + batch_size]
            for i in range(0, len(intermediate_conditions), batch_size)
        ]
        track_chunks = [
            track_conditions[i : i + batch_size]
            for i in range(0, len(track_conditions), batch_size)
        ]

        # Handle the last chunk if it's smaller than 1000 rows
        N_in_final_chunk = int(mother_chunks[-1].shape[0])
        if N_in_final_chunk < batch_size:
            last_chunk = mother_chunks.pop(-1)
            shape = np.asarray(np.shape(last_chunk))
            shape[0] = batch_size
            padded_chunk = np.zeros(tuple(shape))
            padded_chunk[: last_chunk.shape[0]] = last_chunk
            mother_chunks.append(padded_chunk)

            last_chunk = intermediate_chunks.pop(-1)
            shape = np.asarray(np.shape(last_chunk))
            shape[0] = batch_size
            padded_chunk = np.zeros(tuple(shape))
            padded_chunk[: last_chunk.shape[0]] = last_chunk
            intermediate_chunks.append(padded_chunk)

            last_chunk = track_chunks.pop(-1)
            shape = np.asarray(np.shape(last_chunk))
            shape[0] = batch_size
            padded_chunk = np.zeros(tuple(shape))
            padded_chunk[: last_chunk.shape[0]] = last_chunk
            track_chunks.append(padded_chunk)

        return (
            mother_chunks,
            intermediate_chunks,
            track_chunks,
            N_in_final_chunk,
            personalised_transformers,
        )

    def gen_latent(self, network, chunks):
        Nparticles = len(network.branch_options["particles_involved"][0])

        # Generate latent noise tensor
        mother_noise = np.random.normal(
            0,
            1,
            size=(
                np.shape(chunks)[0],
                np.shape(chunks[0])[0],
                network.branch_options["mother_latent_dims"],
            ),
        )

        intermediate_noise = np.random.normal(
            0,
            1,
            size=(
                np.shape(chunks)[0],
                np.shape(chunks[0])[0],
                network.branch_options["intermediate_latent_dims"],
            ),
        )

        track_noise = np.random.normal(
            0,
            1,
            size=(
                np.shape(chunks)[0],
                np.shape(chunks[0])[0],
                Nparticles,
                network.branch_options["track_latent_dims"],
            ),
        )

        return mother_noise, intermediate_noise, track_noise

    def smearPV(self, smeared_PV_output):
        # print("Need to implement function to move the origin vertex too")

        distance_buffer = {}
        for particle in self.particles:
            for coordinate in ["X", "Y", "Z"]:
                distance_buffer[f"{particle}_{coordinate}"] = np.asarray(
                    self.tuple[f"{particle}_orig{coordinate}_TRUE"]
                    - self.tuple[f"{self.mother}_vtx{coordinate}_TRUE"]
                )

        B_plus_TRUEORIGINVERTEX = [
            self.tuple[f"{self.mother}_origX_TRUE"],
            self.tuple[f"{self.mother}_origY_TRUE"],
            self.tuple[f"{self.mother}_origZ_TRUE"],
        ]
        B_plus_TRUEENDVERTEX = [
            self.tuple[f"{self.mother}_vtxX_TRUE"],
            self.tuple[f"{self.mother}_vtxY_TRUE"],
            self.tuple[f"{self.mother}_vtxZ_TRUE"],
        ]
        theta, phi = pts.compute_angles(B_plus_TRUEORIGINVERTEX, B_plus_TRUEENDVERTEX)

        for branch in list(smeared_PV_output.keys()):
            self.tuple[branch] = smeared_PV_output[branch]

        B_plus_TRUEORIGINVERTEX = [
            self.tuple[f"{self.mother}_origX_TRUE"],
            self.tuple[f"{self.mother}_origY_TRUE"],
            self.tuple[f"{self.mother}_origZ_TRUE"],
        ]
        (
            self.tuple[f"{self.mother}_vtxX_TRUE"],
            self.tuple[f"{self.mother}_vtxY_TRUE"],
            self.tuple[f"{self.mother}_vtxZ_TRUE"],
        ) = pts.redefine_endpoint(
            B_plus_TRUEORIGINVERTEX, theta, phi, self.tuple[f"{self.mother}_TRUE_FD"]
        )

        for particle in self.particles:
            for coordinate in ["X", "Y", "Z"]:
                self.tuple[f"{particle}_orig{coordinate}_TRUE"] = (
                    self.tuple[f"{self.mother}_vtx{coordinate}_TRUE"]
                    + distance_buffer[f"{particle}_{coordinate}"]
                    # - distance_buffer[f"{particle}_{coordinate}"] # should be plus
                )

    def updated_smearPV(self, smeared_PV_output):
        # print("Need to implement function to move the origin vertex too")

        for branch in smeared_PV_output:
            self.tuple[branch] = smeared_PV_output[branch]

    def append_trackchi2_conditional_information(self, external_tuple=None, tag=""):
        if external_tuple is not None:
            working_tuple = external_tuple
        else:
            working_tuple = self.tuple

        for idx, particle in enumerate(self.particles):
            working_tuple[f"{tag}{particle}_TRUEID"] = working_tuple[
                f"{particle}_ID_TRUE"
            ]

        for particle in self.particles:
            PT = working_tuple.eval(f"sqrt({particle}_PX**2 + {particle}_PY**2)")

            P = working_tuple.eval(
                f"sqrt({particle}_PX**2 + {particle}_PY**2 + {particle}_PZ**2)"
            )

            working_tuple[f"{tag}{particle}_eta"] = -np.log(
                np.tan(np.arcsin(PT / P) / 2.0)
            )

    def append_initial_conditional_information(
        self, true_PID_scheme, external_tuple=None, tag=""
    ):
        if external_tuple is not None:
            working_tuple = external_tuple
        else:
            working_tuple = self.tuple

        # indices_to_drop = set()  # Store indices to drop later

        for idx, particle in enumerate(self.particles):
            masses = (
                np.ones_like(working_tuple[f"{tag}{particle}_ID_TRUE"])
                * myGlobals.masses[true_PID_scheme[particle]]
            )
            # PIDs_present = []

            # for PID in myGlobals.particle_universe:
            #     where = np.where(
            #         np.abs(working_tuple[f"{tag}{particle}_ID_TRUE"].astype(int)) == PID
            #     )
            #     masses[where] = myGlobals.masses[PID]
            #     if np.shape(where)[1] > 0:
            #         PIDs_present.append(PID)

            # if len(PIDs_present) != 1:
            #     display.warning_print(PIDs_present)
            #     PIDs_present, counts = np.unique(
            #         np.abs(working_tuple[f"{tag}{particle}_ID_TRUE"].astype(int)),
            #         return_counts=True,
            #     )
            #     where = np.where(counts == np.amax(counts))
            #     select = PIDs_present[where][0]

            #     display.warning_print(
            #         "WARNING: particle type cannot change, TRUEID for each final state particle must be the same throughout the tuple."
            #     )
            #     display.warning_print(f"present in {particle} {PIDs_present} (counts: {counts})")
            #     display.warning_print(
            #         f"Marking rows for removal where {tag}{particle}_ID_TRUE != {select}"
            #     )

            #     # Identify indices to drop
            #     drop_indices = working_tuple.index[
            #         np.abs(working_tuple[f"{tag}{particle}_ID_TRUE"]) != select
            #     ].tolist()
            #     indices_to_drop.update(drop_indices)

            # Assign masses after the selection logic
            working_tuple[f"{tag}{particle}_mass"] = masses
            working_tuple[f"{tag}{particle}_E_TRUE"] = pts.compute_TRUEE(
                working_tuple, particle, RapidSim=True
            )
            working_tuple[f"{tag}{particle}_P_TRUE"] = pts.compute_TRUEP(
                working_tuple, particle, RapidSim=True
            )

        # Compute edge angles
        for i, particle_i in enumerate(self.particles):
            for j, particle_j in enumerate(self.particles):
                if i != j:
                    working_tuple[f"edge_angle_{particle_i}_{particle_j}_TRUE"] = (
                        pts.compute_angle(
                            working_tuple,
                            f"{particle_i}",
                            f"{particle_j}",
                            true_vars=True,
                            RapidSim=True,
                        )
                    )

        # # Drop rows at the end
        # if indices_to_drop:
        #     display.warning_print(f"Dropping {len(indices_to_drop)} rows at the end of processing.")
        #     working_tuple = working_tuple.drop(index=indices_to_drop).reset_index(
        #         drop=True
        #     )

    def append_mother_flight(self, external_tuple=None, tag=""):
        if external_tuple is not None:
            working_tuple = external_tuple
        else:
            working_tuple = self.tuple

        mother = "MOTHER"

        A = pts.compute_distance(
            working_tuple, mother, "vtx", mother, "orig", RapidSim=True, true_vars=True
        )
        A = np.asarray(A)
        A[np.where(A == 0)] = 5e-5
        working_tuple[f"{mother}_FLIGHT"] = A

    def append_secondary_conditional_information(self, external_tuple=None, tag=""):
        if external_tuple is not None:
            working_tuple = external_tuple
        else:
            working_tuple = self.tuple

        for particle in self.particles:
            for dim in ["X", "Y", "Z"]:
                working_tuple[f"{particle}_delta_P{dim}"] = (
                    working_tuple[f"{particle}_P{dim}"]
                    - working_tuple[f"{particle}_P{dim}_TRUE"]
                )
            PX = working_tuple[f"{particle}_PX_TRUE"]
            PY = working_tuple[f"{particle}_PY_TRUE"]
            PZ = working_tuple[f"{particle}_PZ_TRUE"]
            PX_reco = working_tuple[f"{particle}_PX"]
            PY_reco = working_tuple[f"{particle}_PY"]
            PZ_reco = working_tuple[f"{particle}_PZ"]

            working_tuple[f"{particle}_P"] = np.sqrt(
                (PX_reco**2 + PY_reco**2 + PZ_reco**2)
            )
            working_tuple[f"{particle}_TRUEP"] = np.sqrt((PX**2 + PY**2 + PZ**2))
            working_tuple[f"{particle}_P_TRUE"] = np.sqrt((PX**2 + PY**2 + PZ**2))
            working_tuple[f"{particle}_Preco_overP"] = (
                working_tuple[f"{particle}_P"] / working_tuple[f"{particle}_TRUEP"]
            )

            working_tuple[f"{particle}_IP"] = pts.compute_impactParameter_i(
                working_tuple,
                "MOTHER",
                f"{particle}",
                true_vertex=True,
                true_vars=False,
                RapidSim=True,
            )
            working_tuple[f"{particle}_IP_TRUE"] = pts.compute_impactParameter_i(
                working_tuple,
                "MOTHER",
                f"{particle}",
                true_vertex=True,
                true_vars=True,
                RapidSim=True,
            )

            working_tuple[f"{particle}_angle_wrt_mother"] = pts.compute_angle(
                working_tuple, "MOTHER", f"{particle}", RapidSim=True
            )
            working_tuple[f"{particle}_angle_wrt_mother_reco"] = pts.compute_angle(
                working_tuple, "MOTHER", f"{particle}", true_vars=False, RapidSim=True
            )

            PT = working_tuple.eval(
                f"sqrt({particle}_PX_TRUE**2 + {particle}_PY_TRUE**2)"
            )
            P = working_tuple.eval(
                f"sqrt({particle}_PX_TRUE**2 + {particle}_PY_TRUE**2 + {particle}_PZ_TRUE**2)"
            )
            working_tuple[f"{particle}_eta_TRUE"] = -np.log(
                np.tan(np.arcsin(PT / P) / 2.0)
            )

            PT = working_tuple.eval(f"sqrt({particle}_PX**2 + {particle}_PY**2)")
            P = working_tuple.eval(
                f"sqrt({particle}_PX**2 + {particle}_PY**2 + {particle}_PZ**2)"
            )
            working_tuple[f"{particle}_eta"] = -np.log(np.tan(np.arcsin(PT / P) / 2.0))

    def append_tertiary_conditional_information(
        self,
        combined_particles,
        true_PID_scheme,
        daughter_particle_names,
        external_tuple=None,
        tag="",
    ):
        if external_tuple is not None:
            working_tuple = external_tuple
        else:
            working_tuple = self.tuple

        mother = "MOTHER"

        # compute distance mother flies before decay
        A = pts.compute_distance(
            working_tuple, mother, "vtx", mother, "orig", RapidSim=True, true_vars=True
        )
        A = np.asarray(A)
        A[np.where(A == 0)] = 5e-5
        working_tuple[f"{mother}_FLIGHT"] = A

        working_tuple[f"{mother}_P_TRUE"] = np.sqrt(
            working_tuple[f"{mother}_PX_TRUE"] ** 2
            + working_tuple[f"{mother}_PY_TRUE"] ** 2
            + working_tuple[f"{mother}_PZ_TRUE"] ** 2
        )
        working_tuple[f"{mother}_PT_TRUE"] = np.sqrt(
            working_tuple[f"{mother}_PX_TRUE"] ** 2
            + working_tuple[f"{mother}_PY_TRUE"] ** 2
        )

        for combined_particle in combined_particles:
            recipe = combined_particles[combined_particle]

            if combined_particle != "MOTHER":
                working_tuple[f"{combined_particle}_PX_TRUE"] = np.zeros(
                    np.shape(working_tuple["MOTHER_PX_TRUE"])
                )
                working_tuple[f"{combined_particle}_PY_TRUE"] = np.zeros(
                    np.shape(working_tuple["MOTHER_PY_TRUE"])
                )
                working_tuple[f"{combined_particle}_PZ_TRUE"] = np.zeros(
                    np.shape(working_tuple["MOTHER_PZ_TRUE"])
                )
                working_tuple[f"{combined_particle}_PX"] = np.zeros(
                    np.shape(working_tuple["MOTHER_PX_TRUE"])
                )
                working_tuple[f"{combined_particle}_PY"] = np.zeros(
                    np.shape(working_tuple["MOTHER_PY_TRUE"])
                )
                working_tuple[f"{combined_particle}_PZ"] = np.zeros(
                    np.shape(working_tuple["MOTHER_PZ_TRUE"])
                )
                for particle in recipe:
                    particle_i = f"DAUGHTER{np.where(np.asarray(daughter_particle_names) == particle)[0][0] + 1}"
                    # print(particle, particle_i)

                    working_tuple[f"{combined_particle}_PX_TRUE"] += working_tuple[
                        f"{particle_i}_PX_TRUE"
                    ]
                    working_tuple[f"{combined_particle}_PY_TRUE"] += working_tuple[
                        f"{particle_i}_PY_TRUE"
                    ]
                    working_tuple[f"{combined_particle}_PZ_TRUE"] += working_tuple[
                        f"{particle_i}_PZ_TRUE"
                    ]
                    working_tuple[f"{combined_particle}_PX"] += working_tuple[
                        f"{particle_i}_PX"
                    ]
                    working_tuple[f"{combined_particle}_PY"] += working_tuple[
                        f"{particle_i}_PY"
                    ]
                    working_tuple[f"{combined_particle}_PZ"] += working_tuple[
                        f"{particle_i}_PZ"
                    ]

        for combined_particle in combined_particles:
            recipe = combined_particles[combined_particle]

            recipe_i = []
            for recipe_ingredient_i in recipe:
                particle_i = f"DAUGHTER{np.where(np.asarray(daughter_particle_names) == recipe_ingredient_i)[0][0] + 1}"
                recipe_i.append(particle_i)

            working_tuple[f"{combined_particle}_DIRA"] = pts.compute_DIRA(
                working_tuple,
                mother,
                recipe_i,
                true_vertex=True,
                true_vars=False,
                RapidSim=True,
            )
            working_tuple[f"{combined_particle}_DIRA_TRUE"] = pts.compute_DIRA(
                working_tuple,
                mother,
                recipe_i,
                true_vertex=True,
                true_vars=True,
                RapidSim=True,
            )
            working_tuple[f"{combined_particle}_IP"] = pts.compute_impactParameter(
                working_tuple,
                mother,
                recipe_i,
                true_vertex=True,
                true_vars=False,
                RapidSim=True,
            )
            working_tuple[f"{combined_particle}_IP_TRUE"] = pts.compute_impactParameter(
                working_tuple,
                mother,
                recipe_i,
                true_vertex=True,
                true_vars=True,
                RapidSim=True,
            )
            (
                working_tuple[f"{combined_particle}_missing_P"],
                working_tuple[f"{combined_particle}_missing_PT"],
            ) = pts.compute_missing_momentum(
                working_tuple, combined_particle, recipe_i, RapidSim=True
            )

        print("WARNING, temporary N_daughters")
        working_tuple["N_daughters"] = 3

        for idx, particle in enumerate(self.particles):
            A = pts.compute_distance(
                working_tuple,
                particle,
                "orig",
                mother,
                "vtx",
                RapidSim=True,
                true_vars=True,
            )
            A = np.asarray(A)
            A[np.where(A == 0)] = 5e-5
            working_tuple[f"{particle}_FLIGHT"] = A

            TRUEID = true_PID_scheme[daughter_particle_names[idx]]
            working_tuple[f"{particle}_TRUEID"] = TRUEID

            working_tuple[f"{particle}_deltaEV_X"] = (
                working_tuple[f"{particle}_origX_TRUE"]
                - working_tuple[f"{mother}_vtxX_TRUE"]
            )
            working_tuple[f"{particle}_deltaEV_Y"] = (
                working_tuple[f"{particle}_origY_TRUE"]
                - working_tuple[f"{mother}_vtxY_TRUE"]
            )
            working_tuple[f"{particle}_deltaEV_Z"] = (
                working_tuple[f"{particle}_origZ_TRUE"]
                - working_tuple[f"{mother}_vtxZ_TRUE"]
            )
            working_tuple[f"{particle}_deltaEV_T"] = np.sqrt(
                working_tuple[f"{particle}_deltaEV_X"] ** 2
                + working_tuple[f"{particle}_deltaEV_Y"] ** 2
            )

            (
                working_tuple[f"{particle}_delta_P"],
                working_tuple[f"{particle}_delta_PT"],
            ) = pts.compute_reconstructed_momentum_residual(
                working_tuple, particle, RapidSim=True
            )

            for dim in ["X", "Y", "Z"]:
                residual_frac = (
                    working_tuple[f"{particle}_P{dim}"]
                    - working_tuple[f"{particle}_P{dim}_TRUE"]
                ) / (working_tuple[f"{particle}_P{dim}_TRUE"] + 1e-4)

                _residualfrac_limit = 5.0
                limit = _residualfrac_limit
                residual_frac[residual_frac < (limit * -1.0)] = -limit
                residual_frac[residual_frac > limit] = limit

                working_tuple[f"{particle}_residualfrac_P{dim}"] = residual_frac



import torch
import torch.nn as nn
import torch.nn.functional as F
from torch_geometric.nn import global_mean_pool, global_add_pool, global_max_pool
from torch_geometric.nn import GATConv, HeteroConv
from torch import Tensor
from typing import Tuple, Optional, List, Dict
import pytorch_lightning as pl

from collections import defaultdict
from rich import print as rprint
from torch_geometric.utils import to_dense_batch

import warnings


from torch_geometric.nn.conv import MessagePassing
from torch_geometric.nn.module_dict import ModuleDict
from torch_geometric.typing import EdgeType, NodeType
from torch_geometric.utils.hetero import check_add_self_loops
from diffusers import DDIMScheduler, DDPMScheduler
import itertools


class myHeteroConv(torch.nn.Module):
    def __init__(
        self,
        convs: Dict[EdgeType, MessagePassing],
        aggr: Optional[str] = "sum",
    ):
        super().__init__()

        for edge_type, module in convs.items():
            check_add_self_loops(module, [edge_type])

        src_node_types = {key[0] for key in convs.keys()}
        dst_node_types = {key[-1] for key in convs.keys()}
        if len(src_node_types - dst_node_types) > 0:
            warnings.warn(
                f"There exist node types ({src_node_types - dst_node_types}) "
                f"whose representations do not get updated during message "
                f"passing as they do not occur as destination type in any "
                f"edge type. This may lead to unexpected behavior."
            )

        self.convs = ModuleDict(convs)
        self.aggr = aggr

    def reset_parameters(self):
        r"""Resets all learnable parameters of the module."""
        for conv in self.convs.values():
            conv.reset_parameters()

    def forward(
        self,
        *args_dict,
        **kwargs_dict,
    ) -> Dict[NodeType, Tensor]:
        out_dict: Dict[str, List[Tensor]] = {}
        if not isinstance(self.aggr, list):
            self.aggr = [self.aggr]

        for edge_type, conv in self.convs.items():
            src, rel, dst = edge_type

            has_edge_level_arg = False

            args = []
            for value_dict in args_dict:
                if edge_type in value_dict:
                    has_edge_level_arg = True
                    args.append(value_dict[edge_type])
                elif src == dst and src in value_dict:
                    args.append(value_dict[src])
                elif src in value_dict or dst in value_dict:
                    args.append(
                        (
                            value_dict.get(src, None),
                            value_dict.get(dst, None),
                        )
                    )

            kwargs = {}
            for arg, value_dict in kwargs_dict.items():
                if not arg.endswith("_dict"):
                    raise ValueError(
                        f"Keyword arguments in '{self.__class__.__name__}' "
                        f"need to end with '_dict' (got '{arg}')"
                    )
                arg = arg[:-5]  # `{*}_dict`
                if edge_type in value_dict:
                    has_edge_level_arg = True
                    kwargs[arg] = value_dict[edge_type]
                elif src == dst and src in value_dict:
                    kwargs[arg] = value_dict[src]
                elif src in value_dict or dst in value_dict:
                    kwargs[arg] = (
                        value_dict.get(src, None),
                        value_dict.get(dst, None),
                    )

            if not has_edge_level_arg:
                continue

            in_values = args_dict[0]
            edge_indicies = args[1]
            out_indices = edge_indicies[1]

            out = conv(*args, **kwargs)

            out_indices = torch.reshape(out_indices, (out.shape[0], out.shape[1]))
            # out_indices = edge_indicies[1]

            if dst not in out_dict:
                out_dict[dst] = [[out, out_indices]]
            else:
                out_dict[dst].append([out, out_indices])

        out_values = {}
        for dst in out_dict:
            for idx, item in enumerate(out_dict[dst]):
                item[0] = torch.reshape(item[0], (-1, item[0].shape[-1]))
                item[1] = torch.flatten(item[1])
                if idx == 0:
                    concat_item = [item[0], item[1]]
                else:
                    concat_item[0] = torch.cat((concat_item[0], item[0]), dim=0)
                    concat_item[1] = torch.cat((concat_item[1], item[1]), dim=0)

            if rel == "down" or self.aggr == ["sum"]:  # no aggr required
                zeros = torch.zeros(in_values[dst].shape, device=out.device)
                out = zeros.scatter_add(
                    0,
                    concat_item[1].unsqueeze(-1).expand_as(concat_item[0]),
                    concat_item[0],
                )
                out_values[dst] = out
            else:
                for idx, aggr in enumerate(self.aggr):
                    if aggr == "sum":
                        zeros = torch.zeros(in_values[dst].shape, device=out.device)
                        aggr_out = zeros.scatter_add(
                            0,
                            concat_item[1].unsqueeze(-1).expand_as(concat_item[0]),
                            concat_item[0],
                        )
                    elif aggr == "mean":
                        zeros = torch.zeros(in_values[dst].shape, device=out.device)
                        sum_out = zeros.scatter_add(
                            0,
                            concat_item[1].unsqueeze(-1).expand_as(concat_item[0]),
                            concat_item[0],
                        )
                        ones = torch.ones_like(concat_item[0])
                        count = zeros.scatter_add(
                            0,
                            concat_item[1].unsqueeze(-1).expand_as(concat_item[0]),
                            ones,
                        )
                        aggr_out = sum_out / count.clamp(min=1)  # Avoid divide-by-zero
                    elif aggr == "max":
                        max_out = torch.zeros_like(zeros)
                        aggr_out = max_out.scatter_reduce(
                            0,
                            concat_item[1].unsqueeze(-1).expand_as(concat_item[0]),
                            concat_item[0],
                            reduce="amax",
                            include_self=True,
                        )

                    if idx == 0:
                        total_aggre = aggr_out
                    else:
                        total_aggre = torch.cat((total_aggre, aggr_out), dim=1)

                aggr_attention = kwargs_dict["aggr_attention_dict"]
                mask = total_aggre.abs().sum(dim=1) == 0
                atten = torch.sigmoid(aggr_attention[f"{dst}_attention"](total_aggre))
                total_aggre = total_aggre * atten  # Avoid `rep[key] *= atten`
                total_aggre = aggr_attention[f"{dst}_aggr"](
                    total_aggre
                )  # Pass through aggregation layer
                total_aggre[mask] = 0

                out_values[dst] = total_aggre

        return out_values

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(num_relations={len(self.convs)})"


class NoAggGAT(MessagePassing):
    def __init__(self, in_channels, out_channels, spec_norm=False):
        super().__init__(aggr=None)  # Keep messages separate
        self.lin = nn.Linear(in_channels, out_channels)
        if spec_norm:
            self.lin = torch.nn.utils.spectral_norm(self.lin)

    def forward(self, x, edge_index):
        # print(x)
        # print(edge_index)
        out = self.propagate(edge_index, x=x)
        # Get unique nodes and counts
        unique_nodes, counts = edge_index[1].unique(return_counts=True)
        max_count = counts.max().item()
        out = out.view(unique_nodes.shape[0], max_count, out.shape[1])

        return out

    # def forward(self, x, edge_index):
    #     out = self.propagate(edge_index, x=x, size=None)

    #     # `index` is the destination node for each message
    #     dst = edge_index[1]
    #     # Group messages by destination node (pad to max)
    #     out, mask = to_dense_batch(out, dst)  # shape: [num_nodes, max_msgs_per_node, out_dim]

    #     return out

    def message(self, x_j):
        return self.lin(x_j)  # Transform node features per edge

    def aggregate(self, inputs, index):
        return inputs  # No aggregation, messages remain separate


def heterogenous_layer_wrapper_w_cat_and_self_loop(
    layer,
    aggr_attention,
    selfloop_layer,
    x,
    edge_index,
    conditions=None,
    infeatures=None,
    aggrs=["sum", "mean", "max"],
):
    edge_index = {key: val for key, val in edge_index.items() if val.numel() != 0}
    # print_heterogenous_representation(x)
    # print('INTO LAYER')
    outputs = layer(x, edge_index, aggr_attention_dict=aggr_attention)

    for node_type in outputs.keys():
        outputs[node_type] = F.leaky_relu(outputs[node_type])

    for node_type in x.keys():
        if (
            node_type in outputs
        ):  # if only some of the nodes of a node type are used as the end point of edges (as defined by edge_index) the un-used are output as 0. - dont want this, so take previous values
            mask = outputs[node_type].abs().sum(dim=1) == 0  # Find zero rows
            concatted_shape = torch.cat([outputs[node_type], x[node_type]], dim=1)
            outputs[node_type] = F.leaky_relu(
                selfloop_layer[node_type](concatted_shape)
            )
            outputs[node_type][mask] = x[node_type][mask]  # Restore original embeddings
        else:
            outputs[node_type] = x[node_type]  # Pass through unchanged

    return outputs


def print_heterogenous_representation(rep):
    for key in rep:
        print(f"{key}:   {rep[key].shape}     (sum: {torch.sum(rep[key])})")
    print("\n")


class TimeEmbedding(nn.Module):
    def __init__(self, embedding_dim):
        super().__init__()
        self.embedding_dim = embedding_dim
        self.linear1 = nn.Linear(1, embedding_dim)
        self.act = nn.SiLU()
        self.linear2 = nn.Linear(embedding_dim, embedding_dim)

    def forward(self, t):
        # t is shape [batch_size], scale to float and unsqueeze
        t = t.unsqueeze(-1).float()
        emb = self.linear1(t)
        emb = self.act(emb)
        emb = self.linear2(emb)
        return emb


class HeteroGenerator(pl.LightningModule):  # 2
    def __init__(
        self,
        track_targets_dims,
        track_conditions_dims,
        hidden_channels,
        track_latent_dims,
        edge_conditions_dims,
        conditonless=False,
    ):
        super(HeteroGenerator, self).__init__()

        self.conditonless = conditonless

        self.time_embedding_dim = 5
        self.time_embed = TimeEmbedding(self.time_embedding_dim)
        self.use_layernorm = True

        # Conditions
        C_track = track_conditions_dims
        if self.conditonless:
            C_track = 0

        # Targets
        T_track = track_targets_dims

        self.add_extra_noise = True
        # Latents
        if self.add_extra_noise:
            # L_track = 3
            # self.L_track = 25
            self.L_track = 3
            # self.L_track = 5
        else:
            self.L_track = 0

        ### ### ### ### ###

        self.aggrs = ["mean", "sum", "max"]

        self.particle_types = ["11", "13", "211", "321", "2212"]
        self.edge_types = list(itertools.product(self.particle_types, repeat=2))

        embedding_dim = 64
        layer = {}
        for particle_type in self.particle_types:
            layer[particle_type] = nn.Linear(T_track + self.L_track, embedding_dim)
            # layer[particle_type] = nn.Sequential(
            #                 nn.Linear(T_track + C_track + L_track, 2 * embedding_dim),
            #                 nn.ReLU(),
            #                 nn.Linear(2 * embedding_dim, embedding_dim)
            #             )
        self.embedding_layer = nn.ModuleDict(layer)
        layer = {}
        for particle_type in self.particle_types:
            layer[particle_type] = nn.Linear(track_conditions_dims, embedding_dim)
            # layer[particle_type] = nn.Sequential(
            #                 nn.Linear(track_conditions_dims, 2 * embedding_dim),
            #                 nn.ReLU(),
            #                 nn.Linear(2 * embedding_dim, embedding_dim)
            #             )
        self.condition_embedding_layer = nn.ModuleDict(layer)
        layer = {}
        for particle_type in self.particle_types:
            layer[particle_type] = nn.Linear(embedding_dim * 2, embedding_dim * 2)
        self.embedding_pooling_attention_layer = nn.ModuleDict(layer)
        layer = {}
        for particle_type in self.particle_types:
            layer[particle_type] = nn.Linear(embedding_dim * 2, embedding_dim)
        self.embedding_pooling_collapsing_layer = nn.ModuleDict(layer)

        # self.communication_dims = 100
        self.communication_dims = hidden_channels

        self.selfloop_layers = nn.ModuleList()
        self.communication_layers = nn.ModuleList()

        self.communication_atten_layers = nn.ModuleList()
        self.communication_aggr_layers = nn.ModuleList()
        self.communication_compressor_layers = nn.ModuleList()
        self.x_compressor_layers = nn.ModuleList()

        self.N_layers = 3
        # self.N_layers = 4
        self.hidden_channels = hidden_channels

        for N_layer in range(self.N_layers):
            layer = {}
            for particle_type in self.particle_types:
                if N_layer == 0:
                    layer[particle_type] = nn.Linear(
                        embedding_dim + self.time_embedding_dim, hidden_channels
                    )
                else:
                    layer[particle_type] = nn.Linear(
                        hidden_channels + self.time_embedding_dim, hidden_channels
                    )
            layer = nn.ModuleDict(layer)
            self.selfloop_layers.append(layer)

            layer_dict = {}
            for edge_type in self.edge_types:
                layer_dict[edge_type[0], "to", edge_type[1]] = GATConv(
                    self.communication_dims,
                    self.communication_dims,
                    add_self_loops=False,
                    edge_dim=edge_conditions_dims,
                )
            layer = HeteroConv(
                layer_dict,
                # aggr="mean", # was max,
                aggr="max",  # was max,
            )
            self.communication_layers.append(layer)

            layer = {}
            for particle_type in self.particle_types:
                layer[particle_type] = nn.Linear(
                    hidden_channels, int(hidden_channels / 2)
                )
            layer = nn.ModuleDict(layer)
            self.communication_compressor_layers.append(layer)
            layer = {}
            for particle_type in self.particle_types:
                layer[particle_type] = nn.Linear(
                    hidden_channels, int(hidden_channels / 2)
                )
            layer = nn.ModuleDict(layer)
            self.x_compressor_layers.append(layer)

            layer = {}
            for particle_type in self.particle_types:
                layer[particle_type] = nn.Linear(hidden_channels, hidden_channels)
            layer = nn.ModuleDict(layer)
            self.communication_atten_layers.append(layer)

            # layer = {}
            # for particle_type in self.particle_types:
            #     layer[particle_type] = nn.Linear(hidden_channels * 2, hidden_channels)
            # layer = nn.ModuleDict(layer)
            # self.communication_aggr_layers.append(layer)

        layer = {}
        for particle_type in self.particle_types:
            layer[particle_type] = nn.Linear(
                hidden_channels + self.time_embedding_dim, hidden_channels
            )
        layer = nn.ModuleDict(layer)
        self.selfloop_layers.append(layer)

        layer = {}
        for particle_type in self.particle_types:
            layer[particle_type] = nn.Linear(
                hidden_channels + self.time_embedding_dim, T_track
            )
        layer = nn.ModuleDict(layer)
        self.selfloop_layers.append(layer)

    def forward(
        self,
        batch,
        timesteps=None,
        add_condition_nodes=False,
        offline_query_mode=False,
    ) -> Tuple[Tensor, Tensor, Tensor]:
        timestep_embedding = self.time_embed(timesteps)

        conditions = {}
        extra_latent = {}
        current_x = {}
        for particle_type in self.particle_types:
            if self.add_extra_noise:
                conditions[particle_type] = (
                    batch[f"{particle_type}_conditions"]
                    .x[:, : -self.L_track]
                    .to(torch.float32)
                )
                extra_latent[particle_type] = (
                    batch[f"{particle_type}_conditions"]
                    .x[:, -self.L_track :]
                    .to(torch.float32)
                )
            else:
                conditions[particle_type] = batch[f"{particle_type}_conditions"].x

            current_x[particle_type] = batch[f"{particle_type}"].x.to(torch.float32)

        edge_index_dict = {}
        edge_attr_dict = {}
        for edge_type in self.edge_types:
            edge_index_dict[edge_type[0], "to", edge_type[1]] = batch[
                edge_type[0], "to", edge_type[1]
            ].edge_index  # ["edge_index"]
            edge_attr_dict[edge_type[0], "to", edge_type[1]] = batch[
                edge_type[0], "to", edge_type[1]
            ].edge_attr  # ["edge_attr"]

        if self.add_extra_noise:
            for node_type in current_x:
                current_x[node_type] = torch.cat(
                    [current_x[node_type], extra_latent[node_type]], dim=1
                )

        x = current_x.copy()

        # if not self.conditonless: # remove conditions, need to keep the extra latent dims that were gracelessly concated to the end of the conditions tensors
        #     for node_type in x:
        #         x[node_type] = torch.cat(
        #             (x[node_type], conditions[node_type]), dim=1
        #         )

        # print(batch)
        for node_type in x:
            # print(node_type, x[node_type].shape)
            if x[node_type].shape[0] > 0:
                x[node_type] = self.embedding_layer[node_type](x[node_type])

        # print('add_condition_nodes', add_condition_nodes)
        # print(x['321'].shape)

        if add_condition_nodes:
            for node_type in x:
                if conditions[node_type].shape[0] > 0:
                    conditions[node_type] = self.condition_embedding_layer[node_type](
                        conditions[node_type]
                    )
                    # print(conditions['321'].shape)
                    x[node_type] = torch.cat(
                        (x[node_type], conditions[node_type]), dim=1
                    )
                    atten = torch.sigmoid(
                        self.embedding_pooling_attention_layer[node_type](x[node_type])
                    )
                    x[node_type] = atten * x[node_type]
                    x[node_type] = self.embedding_pooling_collapsing_layer[node_type](
                        x[node_type]
                    )

        # print(x['321'].shape)

        cat_timesteps = {}
        for node_type in x:
            node_feats = x[node_type]  # [N_nodes, F]
            if conditions[node_type].shape[0] > 0:
                batch_indices = batch[node_type].batch  # [N_nodes]
                node_timesteps = timestep_embedding[
                    batch_indices
                ]  # [N_nodes, D] - select per-node timestep emb
                # Concatenate timestep embedding
                x[node_type] = torch.cat([node_feats, node_timesteps], dim=-1)
                cat_timesteps[node_type] = node_timesteps
            else:
                x[node_type] = torch.cat(
                    [
                        node_feats,
                        torch.empty((0, self.time_embedding_dim)).to(node_feats.device),
                    ],
                    dim=-1,
                )
                cat_timesteps[node_type] = torch.empty((0, self.time_embedding_dim)).to(
                    node_feats.device
                )

        # print(x['321'].shape)
        # print(cat_timesteps['321'].shape)

        x_comms = {}

        for N_layer in range(self.N_layers):
            if N_layer != 0:
                for node_type in x:
                    # print('\n')
                    # print(node_type)
                    # print(x[node_type].shape)
                    # print(conditions[node_type].shape)
                    # print(cat_timesteps[node_type].shape)
                    x[node_type] = torch.cat(
                        [x[node_type], cat_timesteps[node_type]], dim=-1
                    )

            # print(N_layer, x['321'].shape, conditions['321'].shape, cat_timesteps['321'].shape)
            # x_in = {}
            for track_type in x:
                if x[track_type].shape[0] > 0:
                    x[track_type] = F.leaky_relu(
                        self.selfloop_layers[N_layer][track_type](x[track_type])
                    )
                else:
                    x[track_type] = torch.empty((0, self.hidden_channels)).to(
                        x[track_type].device
                    )
            #         x_in[track_type] = x[track_type]
            # x = x_in
            # print_heterogenous_representation(x)

            x_comms = self.communication_layers[N_layer](
                x, edge_index_dict, edge_attr_dict
            )
            for track_type in x_comms:
                if x_comms[track_type].shape[0] > 0:
                    x_comms[track_type] = F.leaky_relu(x_comms[track_type])
                    x[track_type] = torch.cat(
                        [
                            self.x_compressor_layers[N_layer][track_type](
                                x[track_type]
                            ),
                            self.communication_compressor_layers[N_layer][track_type](
                                x_comms[track_type]
                            ),
                        ],
                        dim=-1,
                    )
                    atten = self.communication_atten_layers[N_layer][track_type](
                        x[track_type]
                    )
                    x[track_type] = x[track_type] + torch.sigmoid(atten) * x[track_type]

        for node_type in x:
            x[node_type] = torch.cat([x[node_type], cat_timesteps[node_type]], dim=-1)

        for track_type in x:
            if x[track_type].shape[0] > 0:
                x[track_type] = self.selfloop_layers[-1][track_type](x[track_type])

        return [x]


# class HeteroGenerator(pl.LightningModule):
#     def __init__(
#         self,
#         track_targets_dims,
#         track_conditions_dims,
#         hidden_channels,
#         track_latent_dims,
#         edge_conditions_dims,
#         conditonless=False,

#     ):
#         super(HeteroGenerator, self).__init__()

#         self.conditonless = conditonless

#         self.time_embedding_dim = 5
#         self.time_embed = TimeEmbedding(self.time_embedding_dim)
#         self.use_layernorm = True

#         # Conditions
#         C_track = track_conditions_dims
#         if self.conditonless:
#             C_track = 0

#         # Targets
#         T_track = track_targets_dims

#         self.add_extra_noise = True
#         # Latents
#         if self.add_extra_noise:
#             L_track = 3
#         else: L_track = 0

#         ### ### ### ### ###

#         self.aggrs = ['mean', 'sum','max']

#         self.particle_types = ["11", "13", "211", "321", "2212"]
#         self.edge_types = list(itertools.product(self.particle_types, repeat=2))

#         embedding_dim = 64
#         layer = {}
#         for particle_type in self.particle_types:
#             layer[particle_type] = nn.Linear(T_track + C_track + L_track, embedding_dim)
#             # layer[particle_type] = nn.Sequential(
#             #                 nn.Linear(T_track + C_track + L_track, 2 * embedding_dim),
#             #                 nn.ReLU(),
#             #                 nn.Linear(2 * embedding_dim, embedding_dim)
#             #             )
#         self.embedding_layer = nn.ModuleDict(layer)
#         layer = {}
#         for particle_type in self.particle_types:
#             layer[particle_type] = nn.Linear(track_conditions_dims, embedding_dim)
#             # layer[particle_type] = nn.Sequential(
#             #                 nn.Linear(track_conditions_dims, 2 * embedding_dim),
#             #                 nn.ReLU(),
#             #                 nn.Linear(2 * embedding_dim, embedding_dim)
#             #             )
#         self.condition_embedding_layer = nn.ModuleDict(layer)
#         layer = {}
#         for particle_type in self.particle_types:
#             layer[particle_type] = nn.Linear(embedding_dim*2, embedding_dim*2)
#         self.embedding_pooling_attention_layer = nn.ModuleDict(layer)
#         layer = {}
#         for particle_type in self.particle_types:
#             layer[particle_type] = nn.Linear(embedding_dim*2, embedding_dim)
#         self.embedding_pooling_collapsing_layer = nn.ModuleDict(layer)

#         # self.communication_dims = 100
#         self.communication_dims = hidden_channels

#         self.selfloop_layers = nn.ModuleList()
#         self.communication_layers = nn.ModuleList()

#         self.communication_atten_layers = nn.ModuleList()
#         self.communication_aggr_layers = nn.ModuleList()

#         self.N_layers = 3
#         # self.N_layers = 4

#         for N_layer in range(self.N_layers):

#             layer = {}
#             for particle_type in self.particle_types:
#                 if N_layer == 0:
#                     layer[particle_type] = nn.Linear(embedding_dim + self.time_embedding_dim, hidden_channels)
#                 else:
#                     layer[particle_type] = nn.Linear(hidden_channels, hidden_channels)
#             layer = nn.ModuleDict(layer)
#             self.selfloop_layers.append(layer)

#             layer_dict = {}
#             for edge_type in self.edge_types:
#                 layer_dict[edge_type[0], "to", edge_type[1]] = GATConv(
#                     self.communication_dims,
#                     self.communication_dims,
#                     add_self_loops=False,
#                     edge_dim=edge_conditions_dims,
#                 )
#             layer = HeteroConv(
#                 layer_dict,
#                 aggr="mean", # was max, i
#             )
#             self.communication_layers.append(layer)

#             layer = {}
#             for particle_type in self.particle_types:
#                 layer[particle_type] = nn.Linear(hidden_channels * 2, hidden_channels * 2)
#             layer = nn.ModuleDict(layer)
#             self.communication_atten_layers.append(layer)

#             layer = {}
#             for particle_type in self.particle_types:
#                 layer[particle_type] = nn.Linear(hidden_channels * 2, hidden_channels)
#             layer = nn.ModuleDict(layer)
#             self.communication_aggr_layers.append(layer)


#         layer = {}
#         for particle_type in self.particle_types:
#             layer[particle_type] = nn.Linear(hidden_channels, hidden_channels)
#         layer = nn.ModuleDict(layer)
#         self.selfloop_layers.append(layer)

#         layer = {}
#         for particle_type in self.particle_types:
#             layer[particle_type] = nn.Linear(hidden_channels, T_track)
#         layer = nn.ModuleDict(layer)
#         self.selfloop_layers.append(layer)


#     def forward(
#         self,
#         batch,
#         timesteps=None,
#         add_condition_nodes=False,
#         offline_query_mode=False,
#     ) -> Tuple[Tensor, Tensor, Tensor]:

#         timestep_embedding = self.time_embed(timesteps)

#         conditions = {}
#         extra_latent = {}
#         current_x = {}
#         for particle_type in self.particle_types:
#             # if batch[f"{particle_type}"].x.shape[0] > 0:

#             if self.add_extra_noise:
#                 conditions[particle_type] = batch[f"{particle_type}_conditions"].x[:,:-3].to(torch.float32)
#                 extra_latent[particle_type] = batch[f"{particle_type}_conditions"].x[:,-3:].to(torch.float32)
#             else:
#                 conditions[particle_type] = batch[f"{particle_type}_conditions"].x

#             current_x[particle_type] = batch[f"{particle_type}"].x.to(torch.float32)


#         edge_index_dict = {}
#         edge_attr_dict = {}
#         for edge_type in self.edge_types:
#             edge_index_dict[edge_type[0], "to", edge_type[1]] = batch[
#                 edge_type[0], "to", edge_type[1]
#             ].edge_index  # ["edge_index"]
#             edge_attr_dict[edge_type[0], "to", edge_type[1]] = batch[
#                 edge_type[0], "to", edge_type[1]
#             ].edge_attr  # ["edge_attr"]

#         if self.add_extra_noise:
#             for node_type in current_x:
#                 current_x[node_type] = torch.cat([current_x[node_type], extra_latent[node_type]], dim=1)

#         x = current_x.copy()

#         if not self.conditonless: # remove conditions, need to keep the extra latent dims that were gracelessly concated to the end of the conditions tensors
#             for node_type in x:
#                 x[node_type] = torch.cat(
#                     (x[node_type], conditions[node_type]), dim=1
#                 )

#         for node_type in x:
#             x[node_type] = self.embedding_layer[node_type](x[node_type])

#         if add_condition_nodes:
#             for node_type in x:
#                 if conditions[node_type].shape[0] > 0:
#                     conditions[node_type] = self.condition_embedding_layer[node_type](conditions[node_type])
#                     x[node_type] = torch.cat((x[node_type],conditions[node_type]), dim=1)
#                     atten = torch.sigmoid(self.embedding_pooling_attention_layer[node_type](x[node_type]))
#                     x[node_type] = atten * x[node_type]
#                     x[node_type] = self.embedding_pooling_collapsing_layer[node_type](x[node_type])


#         for node_type in x:
#             node_feats = x[node_type]                      # [N_nodes, F]
#             if conditions[node_type].shape[0] > 0:
#                 batch_indices = batch[node_type].batch              # [N_nodes]
#                 node_timesteps = timestep_embedding[batch_indices]  # [N_nodes, D] - select per-node timestep emb
#                 # Concatenate timestep embedding
#                 x[node_type] = torch.cat([node_feats, node_timesteps], dim=-1)
#             else:
#                 x[node_type] = torch.cat([node_feats, torch.empty((0,self.time_embedding_dim)).to(node_feats.device)], dim=-1)


#         x_comms = {}

#         for N_layer in range(self.N_layers):

#             # residual = {}
#             # if N_layer > 0:
#             #     for track_type in x:
#             #         residual[track_type] = x[track_type]

#             for track_type in x:
#                 x[track_type] = F.leaky_relu(
#                     self.selfloop_layers[N_layer][track_type](x[track_type])
#                 )
#                 # x_comms[track_type] = x[track_type][:, : self.communication_dims]

#             # print_heterogenous_representation(x)
#             # print_heterogenous_representation(x_comms)

#             x_comms = self.communication_layers[N_layer](x, edge_index_dict, edge_attr_dict)
#             for track_type in x_comms:
#                 x_comms[track_type] = F.leaky_relu(x_comms[track_type])
#                 x[track_type] = torch.cat([x[track_type], x_comms[track_type]], dim=-1)
#                 atten = self.communication_atten_layers[N_layer][track_type](x[track_type])
#                 x[track_type] = torch.sigmoid(atten) * x[track_type]
#                 x[track_type] = self.communication_aggr_layers[N_layer][track_type](x[track_type])
#                 x[track_type] = F.leaky_relu(x[track_type])

#             # if N_layer > 0:
#             #     for track_type in x:
#             #         x[track_type] += residual[track_type]


#         for track_type in x:
#             x[track_type] = F.leaky_relu(
#                 self.selfloop_layers[-2][track_type](x[track_type])
#             )

#         for track_type in x:
#             # x[track_type] = torch.tanh(
#             #     self.selfloop_layers[-1][track_type](x[track_type])
#             # )
#             x[track_type] = self.selfloop_layers[-1][track_type](x[track_type])


#         return [x]


import copy


class EMA:
    def __init__(self, model, decay=0.995):
        self.ema_model = copy.deepcopy(model)
        self.ema_model.eval()
        self.decay = decay
        for param in self.ema_model.parameters():
            param.requires_grad = False

    def update(self, model):
        with torch.no_grad():
            msd = model.state_dict()
            for k, ema_v in self.ema_model.state_dict().items():
                model_v = msd[k].detach()  # .to(self.device)
                ema_v.copy_(ema_v * self.decay + (1.0 - self.decay) * model_v)


class diffusion_model_smearing(nn.Module):
    def __init__(
        self,
        track_targets_dims,
        track_conditions_dims,
        hidden_channels,
        track_latent_dims,
        edge_conditions_dims,
        steps,
        silent=False,
        eGAN=False,
    ):
        super(diffusion_model_smearing, self).__init__()

        self.steps = steps

        if not silent:
            print("\n")
            print("Network params -----------------")
            print("track_targets_dims:", track_targets_dims)
            print("track_conditions_dims:", track_conditions_dims)
            print("hidden_channels:", hidden_channels)
            print("track_latent_dims:", track_latent_dims)
            print("edge_conditions_dims:", edge_conditions_dims)
            print("\n")

        self.track_latent_dims = track_latent_dims

        self.track_targets_dims = track_targets_dims

        self.diffuser = HeteroGenerator(
            track_targets_dims,
            track_conditions_dims,
            hidden_channels,
            track_latent_dims,
            edge_conditions_dims,
            conditonless=True,
        )

        self.EMA_diffuser = EMA(self.diffuser)

    def rk4_step(self, f, y, t, dt):
        k1 = f(y, t)
        k2 = f(y + dt * k1 / 2, t + dt / 2)
        k3 = f(y + dt * k2 / 2, t + dt / 2)
        k4 = f(y + dt * k3, t + dt)
        return y + (dt / 6) * (k1 + 2 * k2 + 2 * k3 + k4)

    @torch.no_grad()
    def inference_flow_matching(
        self,
        batch_size,
        batch_in,
        num_steps=100,
        add_condition_nodes=True,
        offline_query_mode=False,
    ):
        print(f"inference_flow_matching, num_steps={num_steps}")

        try:
            batch = batch_in.clone()
        except:
            batch = batch_in

        device = self.diffuser.device
        self.EMA_diffuser.ema_model.to(device)
        particle_types = ["11", "13", "211", "321", "2212"]

        # Valid particle types with data
        valid_particle_types = [
            ptype for ptype in particle_types if batch[ptype].x.shape[0] > 0
        ]

        latent = {}

        for ptype in valid_particle_types:
            n_tracks = batch[ptype].x.shape[0]
            n_conditions = batch[f"{ptype}_conditions"].x.shape[0]
            dtype = batch[f"{ptype}_conditions"].x.dtype

            # Start from Gaussian noise
            latent[ptype] = torch.randn(
                (n_tracks, self.track_latent_dims), device=device
            )
            latent[ptype][:, -2:] = latent[ptype][:, -2:] * 0.25
            # latent[ptype] = 0.25 * torch.randn((n_tracks, self.track_latent_dims), device=device)
            # latent[ptype] = 0.25 * torch.randn((n_tracks, self.track_latent_dims), device=device)
            # latent[ptype] = 0.3 * torch.randn((n_tracks, self.track_latent_dims), device=device)
            # latent[ptype] = 0.2 * torch.randn((n_tracks, self.track_latent_dims), device=device)

            # # Append 3 random noise features to condition
            if self.diffuser.add_extra_noise:
                noise = torch.rand((n_conditions, 25), device=device, dtype=dtype)
                batch[f"{ptype}_conditions"].x = torch.cat(
                    [batch[f"{ptype}_conditions"].x, noise], dim=-1
                )

        # Forward integration from t=0 to t=1
        ts = torch.linspace(0, 1, steps=num_steps + 1, device=device)
        delta_t = ts[1] - ts[0]  # Constant step size

        for i in range(num_steps):
            # # resample noise
            # for ptype in valid_particle_types:
            #     batch[f"{ptype}_conditions"].x[:,-25:] = torch.rand((n_conditions, 25), device=device, dtype=dtype)

            t = ts[i].repeat(batch_size).to(device)

            # Update batch with current latent
            for ptype in valid_particle_types:
                batch[ptype].x = latent[ptype]

            # Predict velocity
            velocity_pred = self.diffuser(
                batch=batch,
                timesteps=t,
                offline_query_mode=offline_query_mode,
                add_condition_nodes=add_condition_nodes,
            )[0]  # Output should be velocity field

            # Euler update
            for ptype in valid_particle_types:
                v = velocity_pred[ptype]
                if v.ndim == 1:
                    v = v.unsqueeze(-1)
                latent[ptype] = latent[ptype] + delta_t * v

            # # RK4 update
            # for ptype in valid_particle_types:
            #     def velocity_fn(y, t_scalar):
            #         # Prepare input batch with y as current latent
            #         batch[ptype].x = y
            #         t_batch = t_scalar.repeat(y.shape[0])

            #         with torch.no_grad():
            #             velocity_pred = self.diffuser(
            #                 batch=batch,
            #                 timesteps=t_batch,
            #                 offline_query_mode=offline_query_mode,
            #                 add_condition_nodes=add_condition_nodes,
            #             )[0]

            #         v = velocity_pred[ptype]
            #         return v if v.ndim > 1 else v.unsqueeze(-1)

            #     latent[ptype] = self.rk4_step(velocity_fn, latent[ptype], t, delta_t)

        return latent

    @torch.no_grad()
    def inference(
        self,
        batch_size,
        batch_in,
        noise_scheduler,
        add_condition_nodes=True,
        offline_query_mode=False,
        CFG=False,
        flowmatching=False,
    ):
        if flowmatching:
            return self.inference_flow_matching(
                batch_size=batch_size,
                batch_in=batch_in,
                # num_steps=100,
                # num_steps=3,
                # num_steps=6,
                # num_steps=10,
                num_steps=25,
                # num_steps=50,
                # num_steps=100,
                add_condition_nodes=add_condition_nodes,
                offline_query_mode=offline_query_mode,
            )

        try:
            batch = batch_in.clone()
        except:
            batch = batch_in

        device = self.diffuser.device
        self.EMA_diffuser.ema_model.to(self.diffuser.device)

        particle_types = ["11", "13", "211", "321", "2212"]
        latent = {}

        # Keep only types with non-zero tracks
        valid_particle_types = [
            ptype for ptype in particle_types if batch[ptype].x.shape[0] > 0
        ]

        for ptype in valid_particle_types:
            n_tracks = batch[ptype].x.shape[0]
            n_conditions = batch[f"{ptype}_conditions"].x.shape[0]
            dtype = batch[f"{ptype}_conditions"].x.dtype

            # Sample latent
            latent[ptype] = torch.randn(
                (n_tracks, self.track_latent_dims), device=device
            )

            # Append 3 random noise features to condition
            noise = torch.rand((n_conditions, 3), device=device, dtype=dtype)
            batch[f"{ptype}_conditions"].x = torch.cat(
                [batch[f"{ptype}_conditions"].x, noise], dim=-1
            )

        # Optionally update `particle_types` to just valid ones
        particle_types = valid_particle_types

        CFG = False
        # CFG = True
        # scale = 1.25  # CFG guidance weight
        scale = 0.75  # CFG guidance weight

        for t in reversed(range(noise_scheduler.config.num_train_timesteps)):
            timesteps = torch.full((batch_size,), t, device=device, dtype=torch.long)

            for particle_type in particle_types:
                batch[particle_type].x = latent[particle_type]

            if CFG:
                # --- 1. Predict noise without condition (unconditional) ---
                uncond_noise_pred = self.diffuser(
                    # uncond_noise_pred = self.EMA_diffuser.ema_model(
                    batch=batch,
                    timesteps=timesteps,
                    offline_query_mode=offline_query_mode,
                    add_condition_nodes=False,  # <-- no conditioning
                )[0]

                # --- 2. Predict noise with condition (conditional) ---
                cond_noise_pred = self.diffuser(
                    # cond_noise_pred = self.EMA_diffuser.ema_model(
                    batch=batch,
                    timesteps=timesteps,
                    offline_query_mode=offline_query_mode,
                    add_condition_nodes=True,  # <-- with conditioning
                )[0]

                # --- 3. Classifier-Free Guidance Combination ---
                guided_noise = {}
                for node_type in latent:
                    cond_out = cond_noise_pred[node_type]
                    uncond_out = uncond_noise_pred[node_type]

                    if cond_out.ndim == 1:
                        cond_out = cond_out.unsqueeze(-1)
                        uncond_out = uncond_out.unsqueeze(-1)

                    guided_noise[node_type] = uncond_out + scale * (
                        cond_out - uncond_out
                    )

                # --- 4. Denoising step ---
                for node_type in latent:
                    model_output = guided_noise[node_type]
                    latent[node_type] = noise_scheduler.step(
                        model_output=model_output, timestep=t, sample=latent[node_type]
                    ).prev_sample

            else:
                # Predict noise at current timestep
                noise_pred = self.diffuser(
                    # noise_pred = self.EMA_diffuser.ema_model(
                    batch=batch,
                    timesteps=timesteps,
                    offline_query_mode=offline_query_mode,
                    add_condition_nodes=add_condition_nodes,
                )[0]

                # Denoise each component
                for node_type in latent:
                    model_output = noise_pred[node_type]

                    # Ensure shape matches expected sample shape
                    if model_output.ndim == 1:
                        model_output = model_output.unsqueeze(
                            -1
                        )  # Or reshape accordingly

                    latent[node_type] = noise_scheduler.step(
                        model_output=model_output, timestep=t, sample=latent[node_type]
                    ).prev_sample

        return latent



import lhcb_rex.settings.globals as myGlobals

import numpy as np
import vector
import uproot


np.seterr(divide="ignore", invalid="ignore")


def write_df_to_root(df, output_name, output_tree):
    df = df.loc[:, ~df.columns.duplicated()]

    branch_dict = {}
    data_dict = {}
    dtypes = df.dtypes
    used_columns = []  # Prevent repeat columns, kpipi_correction was getting repeated

    for branch in df.keys():
        if branch not in used_columns:
            dtype = dtypes[branch]

            # Convert unsigned types to signed
            if dtype == "uint32":
                dtype = "int32"
            elif dtype == "uint64":
                dtype = "int64"

            branch_dict[branch] = dtype

            # Handle DataFrame shapes
            if df[branch].ndim > 1:
                data_dict[branch] = (
                    df[branch].iloc[:, 0].to_numpy()
                )  # Convert to NumPy array
            else:
                data_dict[branch] = df[branch].to_numpy()  # Convert to NumPy array

            if isinstance(data_dict[branch][0], list):
                data_2d = np.vstack(data_dict[branch]).astype(dtype)

                # hack to somehow avoid uproot writing count_branch, n{branch_name}:
                vec = np.empty((np.shape(data_2d)[0], np.shape(data_2d)[1]))
                for dim in range(np.shape(data_2d)[1]):
                    vec[:, dim] = data_2d[:, dim]
                data_dict[branch] = vec

        used_columns.append(branch)

    # Create the ROOT file and write the tree
    with uproot.recreate(output_name) as f:
        # Create a new tree with the specified branches and types
        f[output_tree] = data_dict

    # if myGlobals._verbose:
    # print(f"Tuple written to {output_name}")


def mag(vec):
    sum_sqs = 0
    for component in vec:
        sum_sqs += component**2
    mag = np.sqrt(sum_sqs)
    return mag


def norm(vec):
    mag_vec = mag(vec)

    # mag_vec[np.where(mag_vec==0.)] += 1e-10

    for component_idx in range(np.shape(vec)[0]):
        vec[component_idx] *= 1.0 / mag_vec
    return vec


def dot(vec1, vec2):
    dot = 0
    for component_idx in range(np.shape(vec1)[0]):
        dot += vec1[component_idx] * vec2[component_idx]
    return dot


def compute_angles(origin, end):
    vector = np.array(end) - np.array(origin)
    theta = np.arccos(
        vector[2] / np.sqrt(vector[0] ** 2 + vector[1] ** 2 + vector[2] ** 2)
    )
    phi = np.arctan2(vector[1], vector[0])
    return theta, phi


def redefine_endpoint(origin, theta, phi, new_distance):
    x = new_distance * np.sin(theta) * np.cos(phi)
    y = new_distance * np.sin(theta) * np.sin(phi)
    z = new_distance * np.cos(theta)
    return origin[0] + x, origin[1] + y, origin[2] + z


def compute_distance(df, A, A_var, B, B_var):
    X = df[f"{A}_{A_var}X_TRUE"] - df[f"{B}_{B_var}X_TRUE"]
    Y = df[f"{A}_{A_var}Y_TRUE"] - df[f"{B}_{B_var}Y_TRUE"]
    Z = df[f"{A}_{A_var}Z_TRUE"] - df[f"{B}_{B_var}Z_TRUE"]
    dist = np.sqrt(X**2 + Y**2 + Z**2)
    return dist


def compute_distance_wrapped(df, A, A_var, B, B_var):
    A = compute_distance(df, A, A_var, B, B_var)
    A = np.asarray(A)
    min_A = 5e-5
    A[np.where(A == 0)] = min_A

    return A


def compute_reconstructed_mother_momenta(df, mother, true_vars=True):
    if true_vars:
        PX = df[f"{mother}_PX_TRUE"]
        PY = df[f"{mother}_PY_TRUE"]
        PZ = df[f"{mother}_PZ_TRUE"]
    else:
        PX = df[f"{mother}_PX"]
        PY = df[f"{mother}_PY"]
        PZ = df[f"{mother}_PZ"]

    return np.sqrt((PX**2 + PY**2 + PZ**2)), np.sqrt((PX**2 + PY**2))


def compute_reconstructed_intermediate_momenta(df, particles, true_vars=True):
    PX = 0
    PY = 0
    PZ = 0

    for particle in particles:
        if true_vars:
            PX += df[f"{particle}_PX_TRUE"]
            PY += df[f"{particle}_PY_TRUE"]
            PZ += df[f"{particle}_PZ_TRUE"]
        else:
            PX += df[f"{particle}_PX"]
            PY += df[f"{particle}_PY"]
            PZ += df[f"{particle}_PZ"]

    return np.sqrt((PX**2 + PY**2 + PZ**2)), np.sqrt((PX**2 + PY**2)), PX, PY, PZ


def compute_missing_momentum(df, mother, particles):
    PX = np.asarray(df[f"{mother}_PX_TRUE"]).copy()
    PY = np.asarray(df[f"{mother}_PY_TRUE"]).copy()
    PZ = np.asarray(df[f"{mother}_PZ_TRUE"]).copy()

    for particle in particles:
        PX += -1.0 * np.asarray(df[f"{particle}_PX_TRUE"]).copy()
        PY += -1.0 * np.asarray(df[f"{particle}_PY_TRUE"]).copy()
        PZ += -1.0 * np.asarray(df[f"{particle}_PZ_TRUE"]).copy()

    return np.sqrt((PX**2 + PY**2 + PZ**2)), np.sqrt((PX**2 + PY**2))


def compute_reconstructed_momentum_residual(df, particle):
    PX = df[f"{particle}_PX"] - df[f"{particle}_PX_TRUE"]
    PY = df[f"{particle}_PY"] - df[f"{particle}_PY_TRUE"]
    PZ = df[f"{particle}_PZ"] - df[f"{particle}_PZ_TRUE"]

    return np.sqrt((PX**2 + PY**2 + PZ**2)), np.sqrt((PX**2 + PY**2))


def compute_angle(df, mother, particle, true_vars=True):
    if true_vars:
        momenta_B = np.asarray(
            [df[f"{mother}_PX_TRUE"], df[f"{mother}_PY_TRUE"], df[f"{mother}_PZ_TRUE"]]
        )
        momenta_i = np.asarray(
            [
                df[f"{particle}_PX_TRUE"],
                df[f"{particle}_PY_TRUE"],
                df[f"{particle}_PZ_TRUE"],
            ]
        )
    else:
        momenta_B = np.asarray(
            [df[f"{mother}_PX"], df[f"{mother}_PY"], df[f"{mother}_PZ"]]
        )
        momenta_i = np.asarray(
            [
                df[f"{particle}_PX"],
                df[f"{particle}_PY"],
                df[f"{particle}_PZ"],
            ]
        )

    dot_prod = np.arccos(dot(momenta_B, momenta_i) / (mag(momenta_B) * mag(momenta_i)))

    dot_prod[np.where(np.isnan(dot_prod))] = 1e-6
    dot_prod[np.where(dot_prod == 0)] = 1e-6

    return dot_prod


def compute_impactParameter(df, mother, particles, true_vars=True):
    PX = 0
    PY = 0
    PZ = 0

    for particle in particles:
        if true_vars:
            PX += df[f"{particle}_PX_TRUE"]
            PY += df[f"{particle}_PY_TRUE"]
            PZ += df[f"{particle}_PZ_TRUE"]
        else:
            PX += df[f"{particle}_PX"]
            PY += df[f"{particle}_PY"]
            PZ += df[f"{particle}_PZ"]

    momenta = vector.obj(
        px=PX,
        py=PY,
        pz=PZ,
    )

    end_vertex = np.asarray(
        [
            df[f"{mother}_vtxX_TRUE"],
            df[f"{mother}_vtxY_TRUE"],
            df[f"{mother}_vtxZ_TRUE"],
        ]
    )
    primary_vertex = np.asarray(
        [
            df[f"{mother}_origX_TRUE"],
            df[f"{mother}_origY_TRUE"],
            df[f"{mother}_origZ_TRUE"],
        ]
    )

    momenta_array = np.asarray([momenta.px, momenta.py, momenta.pz])
    P = momenta.mag
    t = 0
    for i in range(3):
        t += momenta_array[i] / P * (primary_vertex[i] - end_vertex[i])
    dist = 0
    for i in range(3):
        dist += (primary_vertex[i] - end_vertex[i] - t * momenta_array[i] / P) ** 2
    dist = np.sqrt(dist)

    return dist


def compute_impactParameter_i(df, mother, particle, true_vars=True):
    if true_vars:
        momenta = vector.obj(
            px=df[f"{particle}_PX_TRUE"],
            py=df[f"{particle}_PY_TRUE"],
            pz=df[f"{particle}_PZ_TRUE"],
        )
    else:
        momenta = vector.obj(
            px=df[f"{particle}_PX"],
            py=df[f"{particle}_PY"],
            pz=df[f"{particle}_PZ"],
        )

    end_vertex = np.asarray(
        [
            df[f"{mother}_vtxX_TRUE"],
            df[f"{mother}_vtxY_TRUE"],
            df[f"{mother}_vtxZ_TRUE"],
        ]
    )
    primary_vertex = np.asarray(
        [
            df[f"{mother}_origX_TRUE"],
            df[f"{mother}_origY_TRUE"],
            df[f"{mother}_origZ_TRUE"],
        ]
    )

    momenta_array = np.asarray([momenta.px, momenta.py, momenta.pz])
    P = momenta.mag
    t = 0
    for i in range(3):
        t += momenta_array[i] / P * (primary_vertex[i] - end_vertex[i])
    dist = 0
    for i in range(3):
        dist += (primary_vertex[i] - end_vertex[i] - t * momenta_array[i] / P) ** 2
    dist = np.sqrt(dist)

    return dist


def compute_flightDistance(df, mother, particles):
    dist = np.sqrt(
        (df[f"{mother}_vtxX_TRUE"] - df[f"{mother}_origX_TRUE"]) ** 2
        + (df[f"{mother}_vtxY_TRUE"] - df[f"{mother}_origY_TRUE"]) ** 2
        + (df[f"{mother}_vtxZ_TRUE"] - df[f"{mother}_origZ_TRUE"]) ** 2
    )

    return dist


def compute_DIRA(df, mother, particles, true_vars=True):
    PX = 0
    PY = 0
    PZ = 0

    for particle in particles:
        if true_vars:
            PX += df[f"{particle}_PX_TRUE"]
            PY += df[f"{particle}_PY_TRUE"]
            PZ += df[f"{particle}_PZ_TRUE"]
        else:
            PX += df[f"{particle}_PX"]
            PY += df[f"{particle}_PY"]
            PZ += df[f"{particle}_PZ"]

    A = norm(np.asarray([PX, PY, PZ]))

    B = norm(
        np.asarray(
            [
                df[f"{mother}_vtxX_TRUE"] - df[f"{mother}_origX_TRUE"],
                df[f"{mother}_vtxY_TRUE"] - df[f"{mother}_origY_TRUE"],
                df[f"{mother}_vtxZ_TRUE"] - df[f"{mother}_origZ_TRUE"],
            ]
        )
    )

    dira = dot(A, B) / np.sqrt(mag(A) ** 2 * mag(B) ** 2)

    return dira



import torch
import torch.nn as nn
import torch.nn.functional as F
from torch_geometric.nn import global_mean_pool, global_add_pool, global_max_pool
from torch_geometric.nn import GATConv, HeteroConv
from torch import Tensor
from typing import Tuple, Optional, List, Dict
import pytorch_lightning as pl

from collections import defaultdict
from rich import print as rprint
from torch_geometric.utils import to_dense_batch

import warnings


from torch_geometric.nn.conv import MessagePassing
from torch_geometric.nn.module_dict import ModuleDict
from torch_geometric.typing import EdgeType, NodeType
from torch_geometric.utils.hetero import check_add_self_loops
from diffusers import DDIMScheduler, DDPMScheduler
import itertools


class myHeteroConv(torch.nn.Module):
    def __init__(
        self,
        convs: Dict[EdgeType, MessagePassing],
        aggr: Optional[str] = "sum",
    ):
        super().__init__()

        for edge_type, module in convs.items():
            check_add_self_loops(module, [edge_type])

        src_node_types = {key[0] for key in convs.keys()}
        dst_node_types = {key[-1] for key in convs.keys()}
        if len(src_node_types - dst_node_types) > 0:
            warnings.warn(
                f"There exist node types ({src_node_types - dst_node_types}) "
                f"whose representations do not get updated during message "
                f"passing as they do not occur as destination type in any "
                f"edge type. This may lead to unexpected behavior."
            )

        self.convs = ModuleDict(convs)
        self.aggr = aggr

    def reset_parameters(self):
        r"""Resets all learnable parameters of the module."""
        for conv in self.convs.values():
            conv.reset_parameters()

    def forward(
        self,
        *args_dict,
        **kwargs_dict,
    ) -> Dict[NodeType, Tensor]:
        out_dict: Dict[str, List[Tensor]] = {}
        if not isinstance(self.aggr, list):
            self.aggr = [self.aggr]

        for edge_type, conv in self.convs.items():
            src, rel, dst = edge_type

            has_edge_level_arg = False

            args = []
            for value_dict in args_dict:
                if edge_type in value_dict:
                    has_edge_level_arg = True
                    args.append(value_dict[edge_type])
                elif src == dst and src in value_dict:
                    args.append(value_dict[src])
                elif src in value_dict or dst in value_dict:
                    args.append(
                        (
                            value_dict.get(src, None),
                            value_dict.get(dst, None),
                        )
                    )

            kwargs = {}
            for arg, value_dict in kwargs_dict.items():
                if not arg.endswith("_dict"):
                    raise ValueError(
                        f"Keyword arguments in '{self.__class__.__name__}' "
                        f"need to end with '_dict' (got '{arg}')"
                    )
                arg = arg[:-5]  # `{*}_dict`
                if edge_type in value_dict:
                    has_edge_level_arg = True
                    kwargs[arg] = value_dict[edge_type]
                elif src == dst and src in value_dict:
                    kwargs[arg] = value_dict[src]
                elif src in value_dict or dst in value_dict:
                    kwargs[arg] = (
                        value_dict.get(src, None),
                        value_dict.get(dst, None),
                    )

            if not has_edge_level_arg:
                continue

            in_values = args_dict[0]
            edge_indicies = args[1]
            out_indices = edge_indicies[1]

            out = conv(*args, **kwargs)

            out_indices = torch.reshape(out_indices, (out.shape[0], out.shape[1]))
            # out_indices = edge_indicies[1]

            if dst not in out_dict:
                out_dict[dst] = [[out, out_indices]]
            else:
                out_dict[dst].append([out, out_indices])

        out_values = {}
        for dst in out_dict:
            for idx, item in enumerate(out_dict[dst]):
                item[0] = torch.reshape(item[0], (-1, item[0].shape[-1]))
                item[1] = torch.flatten(item[1])
                if idx == 0:
                    concat_item = [item[0], item[1]]
                else:
                    concat_item[0] = torch.cat((concat_item[0], item[0]), dim=0)
                    concat_item[1] = torch.cat((concat_item[1], item[1]), dim=0)

            if rel == "down" or self.aggr == ["sum"]:  # no aggr required
                zeros = torch.zeros(in_values[dst].shape, device=out.device)
                out = zeros.scatter_add(
                    0,
                    concat_item[1].unsqueeze(-1).expand_as(concat_item[0]),
                    concat_item[0],
                )
                out_values[dst] = out
            else:
                for idx, aggr in enumerate(self.aggr):
                    if aggr == "sum":
                        zeros = torch.zeros(in_values[dst].shape, device=out.device)
                        aggr_out = zeros.scatter_add(
                            0,
                            concat_item[1].unsqueeze(-1).expand_as(concat_item[0]),
                            concat_item[0],
                        )
                    elif aggr == "mean":
                        zeros = torch.zeros(in_values[dst].shape, device=out.device)
                        sum_out = zeros.scatter_add(
                            0,
                            concat_item[1].unsqueeze(-1).expand_as(concat_item[0]),
                            concat_item[0],
                        )
                        ones = torch.ones_like(concat_item[0])
                        count = zeros.scatter_add(
                            0,
                            concat_item[1].unsqueeze(-1).expand_as(concat_item[0]),
                            ones,
                        )
                        aggr_out = sum_out / count.clamp(min=1)  # Avoid divide-by-zero
                    elif aggr == "max":
                        max_out = torch.zeros_like(zeros)
                        aggr_out = max_out.scatter_reduce(
                            0,
                            concat_item[1].unsqueeze(-1).expand_as(concat_item[0]),
                            concat_item[0],
                            reduce="amax",
                            include_self=True,
                        )

                    if idx == 0:
                        total_aggre = aggr_out
                    else:
                        total_aggre = torch.cat((total_aggre, aggr_out), dim=1)

                aggr_attention = kwargs_dict["aggr_attention_dict"]
                mask = total_aggre.abs().sum(dim=1) == 0
                atten = torch.sigmoid(aggr_attention[f"{dst}_attention"](total_aggre))
                total_aggre = total_aggre * atten  # Avoid `rep[key] *= atten`
                total_aggre = aggr_attention[f"{dst}_aggr"](
                    total_aggre
                )  # Pass through aggregation layer
                total_aggre[mask] = 0

                out_values[dst] = total_aggre

        return out_values

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(num_relations={len(self.convs)})"


class NoAggGAT(MessagePassing):
    def __init__(self, in_channels, out_channels, spec_norm=False):
        super().__init__(aggr=None)  # Keep messages separate
        self.lin = nn.Linear(in_channels, out_channels)
        if spec_norm:
            self.lin = torch.nn.utils.spectral_norm(self.lin)

    def forward(self, x, edge_index):
        # print(x)
        # print(edge_index)
        out = self.propagate(edge_index, x=x)
        # Get unique nodes and counts
        unique_nodes, counts = edge_index[1].unique(return_counts=True)
        max_count = counts.max().item()
        out = out.view(unique_nodes.shape[0], max_count, out.shape[1])

        return out

    # def forward(self, x, edge_index):
    #     out = self.propagate(edge_index, x=x, size=None)

    #     # `index` is the destination node for each message
    #     dst = edge_index[1]
    #     # Group messages by destination node (pad to max)
    #     out, mask = to_dense_batch(out, dst)  # shape: [num_nodes, max_msgs_per_node, out_dim]

    #     return out

    def message(self, x_j):
        return self.lin(x_j)  # Transform node features per edge

    def aggregate(self, inputs, index):
        return inputs  # No aggregation, messages remain separate


def heterogenous_layer_wrapper_w_cat_and_self_loop(
    layer,
    aggr_attention,
    selfloop_layer,
    x,
    edge_index,
    conditions=None,
    infeatures=None,
    aggrs=["sum", "mean", "max"],
):
    edge_index = {key: val for key, val in edge_index.items() if val.numel() != 0}
    # print_heterogenous_representation(x)
    # print('INTO LAYER')
    outputs = layer(x, edge_index, aggr_attention_dict=aggr_attention)

    for node_type in outputs.keys():
        outputs[node_type] = F.leaky_relu(outputs[node_type])

    for node_type in x.keys():
        if (
            node_type in outputs
        ):  # if only some of the nodes of a node type are used as the end point of edges (as defined by edge_index) the un-used are output as 0. - dont want this, so take previous values
            mask = outputs[node_type].abs().sum(dim=1) == 0  # Find zero rows
            concatted_shape = torch.cat([outputs[node_type], x[node_type]], dim=1)
            outputs[node_type] = F.leaky_relu(
                selfloop_layer[node_type](concatted_shape)
            )
            outputs[node_type][mask] = x[node_type][mask]  # Restore original embeddings
        else:
            outputs[node_type] = x[node_type]  # Pass through unchanged

    return outputs


def print_heterogenous_representation(rep):
    for key in rep:
        print(f"{key}:   {rep[key].shape}     (sum: {torch.sum(rep[key])})")
    print("\n")


class TimeEmbedding(nn.Module):
    def __init__(self, embedding_dim):
        super().__init__()
        self.embedding_dim = embedding_dim
        self.linear1 = nn.Linear(1, embedding_dim)
        self.act = nn.SiLU()
        self.linear2 = nn.Linear(embedding_dim, embedding_dim)

    def forward(self, t):
        # t is shape [batch_size], scale to float and unsqueeze
        t = t.unsqueeze(-1).float()
        emb = self.linear1(t)
        emb = self.act(emb)
        emb = self.linear2(emb)
        return emb


class HeteroEncoder(pl.LightningModule):  # 2
    def __init__(
        self,
        track_targets_dims,
        track_conditions_dims,
        hidden_channels,
        track_latent_dims,
        edge_conditions_dims,
        reshape_to_mu_sigma,
        size_of_compressed_space,
        conditonless=False,
    ):
        super(HeteroEncoder, self).__init__()

        self.conditonless = conditonless
        self.reshape_to_mu_sigma = reshape_to_mu_sigma

        self.use_layernorm = True

        # Conditions
        C_track = track_conditions_dims
        if self.conditonless:
            C_track = 0

        # Targets
        T_track = track_targets_dims

        ### ### ### ### ###

        self.aggrs = ["mean", "sum", "max"]

        self.particle_types = ["11", "13", "211", "321", "2212"]
        self.edge_types = list(itertools.product(self.particle_types, repeat=2))

        embedding_dim = 64
        layer = {}
        for particle_type in self.particle_types:
            layer[particle_type] = nn.Linear(T_track + C_track, embedding_dim)
        self.embedding_layer = nn.ModuleDict(layer)
        layer = {}
        for particle_type in self.particle_types:
            layer[particle_type] = nn.Linear(track_conditions_dims, embedding_dim)
        self.condition_embedding_layer = nn.ModuleDict(layer)
        layer = {}
        for particle_type in self.particle_types:
            layer[particle_type] = nn.Linear(embedding_dim * 2, embedding_dim * 2)
        self.embedding_pooling_attention_layer = nn.ModuleDict(layer)
        layer = {}
        for particle_type in self.particle_types:
            layer[particle_type] = nn.Linear(embedding_dim * 2, embedding_dim)
        self.embedding_pooling_collapsing_layer = nn.ModuleDict(layer)

        # self.communication_dims = 100
        self.communication_dims = hidden_channels

        self.selfloop_layers = nn.ModuleList()
        self.communication_layers = nn.ModuleList()

        self.communication_atten_layers = nn.ModuleList()
        self.communication_aggr_layers = nn.ModuleList()
        self.communication_compressor_layers = nn.ModuleList()
        self.x_compressor_layers = nn.ModuleList()

        self.N_layers = 3
        # self.N_layers = 4
        self.hidden_channels = hidden_channels

        for N_layer in range(self.N_layers):
            layer = {}
            for particle_type in self.particle_types:
                if N_layer == 0:
                    layer[particle_type] = nn.Linear(embedding_dim, hidden_channels)
                else:
                    layer[particle_type] = nn.Linear(hidden_channels, hidden_channels)
            layer = nn.ModuleDict(layer)
            self.selfloop_layers.append(layer)

            layer_dict = {}
            for edge_type in self.edge_types:
                layer_dict[edge_type[0], "to", edge_type[1]] = GATConv(
                    self.communication_dims,
                    self.communication_dims,
                    add_self_loops=False,
                    edge_dim=edge_conditions_dims,
                )
            layer = HeteroConv(
                layer_dict,
                # aggr="mean", # was max,
                aggr="max",  # was max,
            )
            self.communication_layers.append(layer)

            layer = {}
            for particle_type in self.particle_types:
                layer[particle_type] = nn.Linear(
                    hidden_channels, int(hidden_channels / 2)
                )
            layer = nn.ModuleDict(layer)
            self.communication_compressor_layers.append(layer)
            layer = {}
            for particle_type in self.particle_types:
                layer[particle_type] = nn.Linear(
                    hidden_channels, int(hidden_channels / 2)
                )
            layer = nn.ModuleDict(layer)
            self.x_compressor_layers.append(layer)

            layer = {}
            for particle_type in self.particle_types:
                layer[particle_type] = nn.Linear(hidden_channels, hidden_channels)
            layer = nn.ModuleDict(layer)
            self.communication_atten_layers.append(layer)

        layer = {}
        for particle_type in self.particle_types:
            layer[particle_type] = nn.Linear(hidden_channels, hidden_channels)
        layer = nn.ModuleDict(layer)
        self.selfloop_layers.append(layer)

        layer = {}
        for particle_type in self.particle_types:
            if self.reshape_to_mu_sigma:
                layer[particle_type] = nn.Linear(
                    hidden_channels, size_of_compressed_space * 2
                )
            else:
                layer[particle_type] = nn.Linear(
                    hidden_channels, size_of_compressed_space
                )

        layer = nn.ModuleDict(layer)
        self.selfloop_layers.append(layer)

    def forward(
        self,
        batch,
        timesteps=None,
        add_condition_nodes=False,
        offline_query_mode=False,
    ) -> Tuple[Tensor, Tensor, Tensor]:
        conditions = {}
        extra_latent = {}
        current_x = {}
        for particle_type in self.particle_types:
            conditions[particle_type] = batch[f"{particle_type}_conditions"].x
            current_x[particle_type] = batch[f"{particle_type}"].x.to(torch.float32)

        edge_index_dict = {}
        edge_attr_dict = {}
        for edge_type in self.edge_types:
            edge_index_dict[edge_type[0], "to", edge_type[1]] = batch[
                edge_type[0], "to", edge_type[1]
            ].edge_index  # ["edge_index"]
            edge_attr_dict[edge_type[0], "to", edge_type[1]] = batch[
                edge_type[0], "to", edge_type[1]
            ].edge_attr  # ["edge_attr"]

        x = current_x.copy()

        if not self.conditonless:  # remove conditions, need to keep the extra latent dims that were gracelessly concated to the end of the conditions tensors
            for node_type in x:
                x[node_type] = torch.cat((x[node_type], conditions[node_type]), dim=1)

        # print(batch)
        for node_type in x:
            # print(node_type, x[node_type].shape)
            if x[node_type].shape[0] > 0:
                x[node_type] = self.embedding_layer[node_type](x[node_type])

        if add_condition_nodes:
            for node_type in x:
                if conditions[node_type].shape[0] > 0:
                    conditions[node_type] = self.condition_embedding_layer[node_type](
                        conditions[node_type]
                    )
                    x[node_type] = torch.cat(
                        (x[node_type], conditions[node_type]), dim=1
                    )
                    atten = torch.sigmoid(
                        self.embedding_pooling_attention_layer[node_type](x[node_type])
                    )
                    x[node_type] = atten * x[node_type]
                    x[node_type] = self.embedding_pooling_collapsing_layer[node_type](
                        x[node_type]
                    )

        x_comms = {}

        for N_layer in range(self.N_layers):
            # x_in = {}
            for track_type in x:
                if x[track_type].shape[0] > 0:
                    x[track_type] = F.leaky_relu(
                        self.selfloop_layers[N_layer][track_type](x[track_type])
                    )
                else:
                    x[track_type] = torch.empty((0, self.hidden_channels)).to(
                        x[track_type].device
                    )

            # print_heterogenous_representation(x)

            x_comms = self.communication_layers[N_layer](
                x, edge_index_dict, edge_attr_dict
            )
            for track_type in x_comms:
                if x_comms[track_type].shape[0] > 0:
                    x_comms[track_type] = F.leaky_relu(x_comms[track_type])
                    x[track_type] = torch.cat(
                        [
                            self.x_compressor_layers[N_layer][track_type](
                                x[track_type]
                            ),
                            self.communication_compressor_layers[N_layer][track_type](
                                x_comms[track_type]
                            ),
                        ],
                        dim=-1,
                    )
                    atten = self.communication_atten_layers[N_layer][track_type](
                        x[track_type]
                    )
                    x[track_type] = x[track_type] + torch.sigmoid(atten) * x[track_type]

        for track_type in x:
            if x[track_type].shape[0] > 0:
                x[track_type] = self.selfloop_layers[-1][track_type](x[track_type])
                if self.reshape_to_mu_sigma:
                    x[track_type] = x[track_type].view(
                        -1, int(x[track_type].shape[-1] / 2), 2
                    )
        return [x]


import copy


class EMA:
    def __init__(self, model, decay=0.995):
        self.ema_model = copy.deepcopy(model)
        self.ema_model.eval()
        self.decay = decay
        for param in self.ema_model.parameters():
            param.requires_grad = False

    def update(self, model):
        with torch.no_grad():
            msd = model.state_dict()
            for k, ema_v in self.ema_model.state_dict().items():
                model_v = msd[k].detach()  # .to(self.device)
                ema_v.copy_(ema_v * self.decay + (1.0 - self.decay) * model_v)


class VAE_smearing(nn.Module):
    def __init__(
        self,
        track_targets_dims,
        track_conditions_dims,
        hidden_channels,
        track_latent_dims,
        edge_conditions_dims,
        steps,
        silent=False,
        eGAN=False,
    ):
        super(VAE_smearing, self).__init__()

        self.steps = steps

        if not silent:
            print("\n")
            print("Network params -----------------")
            print("track_targets_dims:", track_targets_dims)
            print("track_conditions_dims:", track_conditions_dims)
            print("hidden_channels:", hidden_channels)
            print("track_latent_dims:", track_latent_dims)
            print("edge_conditions_dims:", edge_conditions_dims)
            print("\n")

        self.track_latent_dims = track_latent_dims

        self.track_targets_dims = track_targets_dims

        self.size_of_compressed_space = 3

        self.encoder = HeteroEncoder(
            track_targets_dims,
            track_conditions_dims,
            hidden_channels,
            track_latent_dims,
            edge_conditions_dims,
            reshape_to_mu_sigma=True,
            size_of_compressed_space=self.size_of_compressed_space,
            conditonless=True,
        )

        self.decoder = HeteroEncoder(
            self.size_of_compressed_space,
            track_conditions_dims,
            hidden_channels,
            track_latent_dims,
            edge_conditions_dims,
            reshape_to_mu_sigma=False,
            size_of_compressed_space=track_targets_dims,
            conditonless=True,
        )

    def stacked(self, batch_in):
        try:
            batch = batch_in.clone()
        except:
            batch = batch_in

        latent = self.encoder(
            batch=batch,
        )[0]

        particle_types = ["11", "13", "211", "321", "2212"]
        valid_particle_types = [
            ptype for ptype in particle_types if batch[ptype].x.shape[0] > 0
        ]

        mu = {}
        logvar = {}

        for ptype in valid_particle_types:
            # print('reparam')
            mu_i = latent[ptype][:, :, 0]
            logvar_i = latent[ptype][:, :, 1]

            mu[ptype] = mu_i
            logvar[ptype] = logvar_i

            eps = torch.randn_like(logvar_i)
            std = torch.exp(0.5 * logvar_i)
            z = mu_i + std * eps

            batch[ptype].x = z

        out = self.decoder(
            batch=batch,
        )[0]

        return out, mu, logvar

    @torch.no_grad()
    def inference(self, batch_size, batch_in, offline_query_mode=False):
        try:
            batch = batch_in.clone()
        except:
            batch = batch_in

        # device = self.decoder.device

        # particle_types = ["11", "13", "211", "321", "2212"]

        # # Keep only types with non-zero tracks
        # valid_particle_types = [ptype for ptype in particle_types if batch[ptype].x.shape[0] > 0]

        # for ptype in valid_particle_types:
        #     n_tracks = batch[ptype].x.shape[0]

        #     # Sample latent
        #     batch[ptype].x = torch.randn((n_tracks, self.size_of_compressed_space), device=device)

        # # Predict noise at current timestep
        # out = self.decoder(
        #     batch=batch,
        # )[0]

        out, mu, sigma = self.stacked(batch)

        return out
